#!/usr/bin/env python
# coding: utf-8 -*-
"""
ANTA Top-level Console
https://rich.readthedocs.io/en/stable/console.html#console-api
"""

from rich.console import Console
from rich.theme import Theme

from anta import RICH_COLOR_THEME

console = Console(theme=Theme(RICH_COLOR_THEME))



"""
Toolkit for ANTA to play with Pydantic.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, List, Sequence

if TYPE_CHECKING:
    from anta.result_manager.models import ListResult

logger = logging.getLogger(__name__)


def pydantic_to_dict(pydantic_list: ListResult) -> List[Dict[str, Sequence[Any]]]:
    """
    Convert Pydantic object into a list of dict

    Mimic .dict() option from pydantic but overwrite IPv4Address nodes

    Args:
        pydantic_list: Iterable pydantic object

    Returns:
        List[Dict[str, str]]: the list of dict
    """
    result = []
    for device in pydantic_list:
        dev_dict = {k: str(v) if not isinstance(v, list) else v for k, v in device}
        result.append(dev_dict)
    return result



"""Models related to anta.result_manager module."""

from pydantic import BaseModel, validator
from rich.text import Text

from ..result_manager.models import RESULT_OPTIONS


class ColorManager(BaseModel):
    """Color management for status report.

    Attributes:
        level (str): Test result value.
        color (str): Associated color.
    """

    level: str
    color: str

    @validator("level", allow_reuse=True)
    def name_must_be_in(cls, v: str) -> str:
        """
        Status validator

        Validate status is a supported one

        Args:
            v (str): User defined level

        Raises:
            ValueError: If level is unsupported

        Returns:
            str: level value
        """
        if v not in RESULT_OPTIONS:
            raise ValueError(f"must be one of {RESULT_OPTIONS}")
        return v

    def style_rich(self) -> Text:
        """
        Build a rich Text syntax with color

        Returns:
            Text: object with level string and its associated color.
        """
        return Text(self.level, style=self.color)

    def string(self) -> str:
        """
        Build an str with color code

        Returns:
            str: String with level and its associated color
        """
        return f"[{self.color}]{self.level}"



"""
Test functions related to VXLAN
"""

from anta.models import AntaCommand, AntaTest


class VerifyVxlan1Interface(AntaTest):
    """
    This test verifies if the Vxlan1 interface is configured and 'up/up'.

    !!! warning
        The name of this test has been updated from 'VerifyVxlan' for better representation.

    Expected Results:
      * success: The test will pass if the Vxlan1 interface is configured with line protocol status and interface status 'up'.
      * failure: The test will fail if the Vxlan1 interface line protocol status or interface status are not 'up'.
      * skipped: The test will be skipped if the Vxlan1 interface is not configured.
    """

    name = "VerifyVxlan1Interface"
    description = "This test verifies if the Vxlan1 interface is configured and 'up/up'."
    categories = ["vxlan"]
    commands = [AntaCommand(command="show interfaces description", ofmt="json")]

    @AntaTest.anta_test
    def test(self) -> None:
        """
        Run VerifyVxlan1Interface validation
        """

        command_output = self.instance_commands[0].json_output

        if "Vxlan1" not in command_output["interfaceDescriptions"]:
            self.result.is_skipped("Vxlan1 interface is not configured")
        elif (
            command_output["interfaceDescriptions"]["Vxlan1"]["lineProtocolStatus"] == "up"
            and command_output["interfaceDescriptions"]["Vxlan1"]["interfaceStatus"] == "up"
        ):
            self.result.is_success()
        else:
            self.result.is_failure(
                f"Vxlan1 interface is {command_output['interfaceDescriptions']['Vxlan1']['lineProtocolStatus']}"
                f"/{command_output['interfaceDescriptions']['Vxlan1']['interfaceStatus']}"
            )


class VerifyVxlanConfigSanity(AntaTest):
    """
    This test verifies that no issues are detected with the VXLAN configuration.

    Expected Results:
      * success: The test will pass if no issues are detected with the VXLAN configuration.
      * failure: The test will fail if issues are detected with the VXLAN configuration.
      * skipped: The test will be skipped if VXLAN is not configured on the device.
    """

    name = "VerifyVxlanConfigSanity"
    description = "This test verifies that no issues are detected with the VXLAN configuration."
    categories = ["vxlan"]
    commands = [AntaCommand(command="show vxlan config-sanity", ofmt="json")]

    @AntaTest.anta_test
    def test(self) -> None:
        """
        Run VerifyVxlanConfigSanity validation
        """

        command_output = self.instance_commands[0].json_output

        if "categories" not in command_output or len(command_output["categories"]) == 0:
            self.result.is_skipped("VXLAN is not configured")
            return

        failed_categories = {
            category: content
            for category, content in command_output["categories"].items()
            if category in ["localVtep", "mlag", "pd"] and content["allCheckPass"] is not True
        }

        if len(failed_categories) > 0:
            self.result.is_failure(f"VXLAN config sanity check is not passing: {failed_categories}")
        else:
            self.result.is_success()



"""
Test functions related to various Spanning Tree Protocol (STP) settings
"""
from __future__ import annotations

from typing import List, Optional

from anta.models import AntaCommand, AntaTemplate, AntaTest
from anta.tools.get_value import get_value


class VerifySTPMode(AntaTest):
    """
    Verifies the configured STP mode for a provided list of VLAN(s).

    Expected Results:
        * success: The test will pass if the STP mode is configured properly in the specified VLAN(s).
        * failure: The test will fail if the STP mode is NOT configured properly for one or more specified VLAN(s).
        * skipped: The test will be skipped if the STP mode is not provided.
        * error: The test will give an error if a list of VLAN(s) is not provided as template_params.
    """

    name = "VerifySTPMode"
    description = "Verifies the configured STP mode for a provided list of VLAN(s)."
    categories = ["stp"]
    template = AntaTemplate(template="show spanning-tree vlan {vlan}")

    @staticmethod
    def _check_stp_mode(mode: str) -> None:
        """
        Verifies if the provided STP mode is compatible with Arista EOS devices.

        Args:
            mode: The STP mode to verify.
        """
        stp_modes = ["mstp", "rstp", "rapidPvst"]

        if mode not in stp_modes:
            raise ValueError(f"Wrong STP mode provided. Valid modes are: {stp_modes}")

    @AntaTest.anta_test
    def test(self, mode: str = "mstp") -> None:
        """
        Run VerifySTPVersion validation.

        Args:
            mode: STP mode to verify. Defaults to 'mstp'.
        """
        if not mode:
            self.result.is_skipped(f"{self.__class__.name} did not run because mode was not supplied")
            return

        self._check_stp_mode(mode)

        self.result.is_success()

        for command in self.instance_commands:
            if command.params and "vlan" in command.params:
                vlan_id = command.params["vlan"]
            if not (stp_mode := get_value(command.json_output, f"spanningTreeVlanInstances.{vlan_id}.spanningTreeVlanInstance.protocol")):
                self.result.is_failure(f"STP mode '{mode}' not configured for VLAN {vlan_id}")

            elif stp_mode != mode:
                self.result.is_failure(f"Wrong STP mode configured for VLAN {vlan_id}")


class VerifySTPBlockedPorts(AntaTest):
    """
    Verifies there is no STP blocked ports.

    Expected Results:
        * success: The test will pass if there are NO ports blocked by STP.
        * failure: The test will fail if there are ports blocked by STP.
    """

    name = "VerifySTPBlockedPorts"
    description = "Verifies there is no STP blocked ports."
    categories = ["stp"]
    commands = [AntaCommand(command="show spanning-tree blockedports")]

    @AntaTest.anta_test
    def test(self) -> None:
        """
        Run VerifySTPBlockedPorts validation
        """

        command_output = self.instance_commands[0].json_output

        if not (stp_instances := command_output["spanningTreeInstances"]):
            self.result.is_success()
        else:
            for key, value in stp_instances.items():
                stp_instances[key] = value.pop("spanningTreeBlockedPorts")
            self.result.is_failure(f"The following ports are blocked by STP: {stp_instances}")


class VerifySTPCounters(AntaTest):
    """
    Verifies there is no errors in STP BPDU packets.

    Expected Results:
        * success: The test will pass if there are NO STP BPDU packet errors under all interfaces participating in STP.
        * failure: The test will fail if there are STP BPDU packet errors on one or many interface(s).
    """

    name = "VerifySTPCounters"
    description = "Verifies there is no errors in STP BPDU packets."
    categories = ["stp"]
    commands = [AntaCommand(command="show spanning-tree counters")]

    @AntaTest.anta_test
    def test(self) -> None:
        """
        Run VerifySTPBlockedPorts validation
        """

        command_output = self.instance_commands[0].json_output

        interfaces_with_errors = [
            interface for interface, counters in command_output["interfaces"].items() if counters["bpduTaggedError"] or counters["bpduOtherError"] != 0
        ]

        if interfaces_with_errors:
            self.result.is_failure(f"The following interfaces have STP BPDU packet errors: {interfaces_with_errors}")
        else:
            self.result.is_success()


class VerifySTPForwardingPorts(AntaTest):
    """
    Verifies that all interfaces are in a forwarding state for a provided list of VLAN(s).

    Expected Results:
        * success: The test will pass if all interfaces are in a forwarding state for the specified VLAN(s).
        * failure: The test will fail if one or many interfaces are NOT in a forwarding state in the specified VLAN(s).
        * error: The test will give an error if a list of VLAN(s) is not provided as template_params.
    """

    name = "VerifySTPForwardingPorts"
    description = "Verifies that all interfaces are forwarding for a provided list of VLAN(s)."
    categories = ["stp"]
    template = AntaTemplate(template="show spanning-tree topology vlan {vlan} status")

    @AntaTest.anta_test
    def test(self) -> None:
        """
        Run VerifySTPForwardingPorts validation.
        """

        self.result.is_success()

        for command in self.instance_commands:
            if command.params and "vlan" in command.params:
                vlan_id = command.params["vlan"]

            if not (topologies := get_value(command.json_output, "topologies")):
                self.result.is_failure(f"STP instance for VLAN {vlan_id} is not configured")

            else:
                for value in topologies.values():
                    if int(vlan_id) in value["vlans"]:
                        interfaces_not_forwarding = [interface for interface, state in value["interfaces"].items() if state["state"] != "forwarding"]

                if interfaces_not_forwarding:
                    self.result.is_failure(f"The following interface(s) are not in a forwarding state for VLAN {vlan_id}: {interfaces_not_forwarding}")


class VerifySTPRootPriority(AntaTest):
    """
    Verifies the STP root priority for a provided list of VLAN or MST instance ID(s).

    Expected Results:
        * success: The test will pass if the STP root priority is configured properly for the specified VLAN or MST instance ID(s).
        * failure: The test will fail if the STP root priority is NOT configured properly for the specified VLAN or MST instance ID(s).
        * skipped: The test will be skipped if the STP root priority is not provided.
    """

    name = "VerifySTPRootPriority"
    description = "Verifies the STP root priority for a provided list of VLAN or MST instance ID(s)."
    categories = ["stp"]
    commands = [AntaCommand(command="show spanning-tree root detail")]

    @AntaTest.anta_test
    def test(self, priority: Optional[int] = None, instances: Optional[List[int]] = None) -> None:
        """
        Run VerifySTPRootPriority validation.

        Args:
            priority: STP root priority to verify.
            instances: List of VLAN or MST instance ID(s). By default, ALL VLAN or MST instance ID(s) will be verified.
        """
        if not priority:
            self.result.is_skipped(f"{self.__class__.name} did not run because priority was not supplied")
            return

        command_output = self.instance_commands[0].json_output

        if not (stp_instances := command_output["instances"]):
            self.result.is_failure("No STP instances configured")
            return

        for instance in stp_instances:
            if instance.startswith("MST"):
                prefix = "MST"
                break
            if instance.startswith("VL"):
                prefix = "VL"
                break

        check_instances = [f"{prefix}{instance_id}" for instance_id in instances] if instances else command_output["instances"].keys()

        wrong_priority_instances = [instance for instance in check_instances if get_value(command_output, f"instances.{instance}.rootBridge.priority") != priority]

        if wrong_priority_instances:
            self.result.is_failure(f"The following instance(s) have the wrong STP root priority configured: {wrong_priority_instances}")
        else:
            self.result.is_success()



"""
Test functions related to multicast
"""

from typing import List, Optional

from anta.models import AntaCommand, AntaTest


class VerifyIGMPSnoopingVlans(AntaTest):
    """
    Verifies the IGMP snooping configuration for some VLANs.

    Args:
        vlans (List[str]): A list of VLANs
        configuration (str): Expected IGMP snooping configuration (enabled or disabled) for these VLANs.
    """

    name = "VerifyIGMPSnoopingVlans"
    description = "Verifies the IGMP snooping configuration for some VLANs."
    categories = ["multicast", "igmp"]
    commands = [AntaCommand(command="show ip igmp snooping")]

    @AntaTest.anta_test
    def test(self, vlans: Optional[List[str]] = None, configuration: Optional[str] = None) -> None:
        """
        Run VerifyIGMPSnoopingVlans validation

        Args:
            vlans: List of VLANs.
            configuration: Expected IGMP configuration (enabled or disabled) for these VLANs.
        """

        if not vlans or not configuration:
            self.result.is_skipped("VerifyIGMPSnoopingVlans was not run as no vlans or configuration was given")
            return
        if configuration not in ["enabled", "disabled"]:
            self.result.is_error(f"VerifyIGMPSnoopingVlans was not run as 'configuration': {configuration} is not in the allowed values: ['enabled', 'disabled'])")
            return

        command_output = self.instance_commands[0].json_output

        self.result.is_success()
        for vlan in vlans:
            if vlan not in command_output["vlans"]:
                self.result.is_failure(f"Supplied vlan {vlan} is not present on the device.")
                continue

            igmp_state = command_output["vlans"][str(vlan)]["igmpSnoopingState"]
            if igmp_state != configuration:
                self.result.is_failure(f"IGMP state for vlan {vlan} is {igmp_state}")


class VerifyIGMPSnoopingGlobal(AntaTest):
    """
    Verifies the IGMP snooping global configuration.

    Args:
        configuration (str): Expected global IGMP snooping configuration (enabled or disabled).
    """

    name = "VerifyIGMPSnoopingGlobal"
    description = "Verifies the IGMP snooping global configuration."
    categories = ["multicast", "igmp"]
    commands = [AntaCommand(command="show ip igmp snooping")]

    @AntaTest.anta_test
    def test(self, configuration: Optional[str] = None) -> None:
        """
        Run VerifyIGMPSnoopingGlobal validation

        Args:
            configuration: Expected global IGMP configuration (enabled or disabled).
        """

        if not configuration:
            self.result.is_skipped("VerifyIGMPSnoopingGlobal was not run as no configuration was given")
            return

        if configuration not in ["enabled", "disabled"]:
            self.result.is_error(f"VerifyIGMPSnoopingGlobal was not run as 'configuration': {configuration} is not in the allowed values: ['enabled', 'disabled'])")
            return

        command_output = self.instance_commands[0].json_output

        self.result.is_success()
        if (igmp_state := command_output["igmpSnoopingState"]) != configuration:
            self.result.is_failure(f"IGMP state is not valid: {igmp_state}")



"""
Test functions related to various connectivity checks
"""
from __future__ import annotations

from anta.models import AntaTemplate, AntaTest


class VerifyReachability(AntaTest):
    """
    Test network reachability to one or many destination IP(s).

    Expected Results:
        * success: The test will pass if all destination IP(s) are reachable.
        * failure: The test will fail if one or many destination IP(s) are unreachable.
        * error: The test will give an error if the destination IP(s) or the source interface/IP(s) are not provided as template_params.
    """

    name = "VerifyReachability"
    description = "Test the network reachability to one or many destination IP(s)."
    categories = ["connectivity"]
    template = AntaTemplate(template="ping {dst} source {src} repeat 2")

    @AntaTest.anta_test
    def test(self) -> None:
        """
        Run VerifyReachability validation.
        """

        failures = []

        for command in self.instance_commands:
            if command.params and ("src" and "dst") in command.params:
                src, dst = command.params["src"], command.params["dst"]
            else:
                self.result.is_error("The destination IP(s) or the source interface/IP(s) are not provided as template_params")
                return

            if "2 received" not in command.json_output["messages"][0]:
                failures.append((src, dst))

        if not failures:
            self.result.is_success()

        else:
            self.result.is_failure(f"Connectivity test failed for the following source-destination pairs: {failures}")



#!/usr/bin/env python
# coding: utf-8 -*-
"""
ANTA CLI
"""

import logging
import pathlib
from typing import Any, Callable, Dict, List, Tuple

import click

from anta import __version__
from anta.cli.debug import commands as debug_commands
from anta.cli.exec import commands as exec_commands
from anta.cli.get import commands as get_commands
from anta.cli.nrfu import commands as check_commands
from anta.cli.utils import IgnoreRequiredWithHelp, parse_catalog, parse_inventory, requires_enable, setup_logging
from anta.result_manager.models import TestResult


# @click.group()
@click.group(cls=IgnoreRequiredWithHelp)
@click.pass_context
@click.version_option(__version__)
@click.option(
    "--username",
    show_envvar=True,
    help="Username to connect to EOS",
    required=True,
)
@click.option(
    "--password",
    show_envvar=True,
    help="Password to connect to EOS",
    required=True,
)
@click.option(
    "--timeout",
    show_envvar=True,
    default=5,
    help="Global connection timeout",
    show_default=True,
)
@click.option(
    "--insecure",
    show_envvar=True,
    is_flag=True,
    default=False,
    help="Disable SSH Host Key validation",
    show_default=True,
)
@click.option(
    "--enable",
    show_envvar=True,
    is_flag=True,
    default=False,
    help="Add enable mode towards the devices if required to connect",
    show_default=True,
)
@click.option(
    "--enable-password",
    show_envvar=True,
    help="Enable password if required to connect, --enable MUST be set",
    callback=requires_enable,
)
@click.option(
    "--inventory",
    "-i",
    show_envvar=True,
    required=True,
    help="Path to the inventory YAML file",
    type=click.Path(file_okay=True, dir_okay=False, exists=True, readable=True, path_type=pathlib.Path),
)
@click.option(
    "--log-level",
    "--log",
    show_envvar=True,
    help="ANTA logging level",
    default=logging.getLevelName(logging.INFO),
    show_default=True,
    type=click.Choice(
        [
            logging.getLevelName(logging.CRITICAL),
            logging.getLevelName(logging.ERROR),
            logging.getLevelName(logging.WARNING),
            logging.getLevelName(logging.INFO),
            logging.getLevelName(logging.DEBUG),
        ],
        case_sensitive=False,
    ),
    callback=setup_logging,
)
@click.option("--ignore-status", show_envvar=True, is_flag=True, default=False, help="Always exit with success")
@click.option("--ignore-error", show_envvar=True, is_flag=True, default=False, help="Only report failures and not errors")
def anta(ctx: click.Context, inventory: pathlib.Path, ignore_status: bool, ignore_error: bool, **kwargs: Any) -> None:
    # pylint: disable=unused-argument
    """Arista Network Test Automation (ANTA) CLI"""
    ctx.ensure_object(dict)
    ctx.obj["inventory"] = parse_inventory(ctx, inventory)
    ctx.obj["ignore_status"] = ignore_status
    ctx.obj["ignore_error"] = ignore_error


@anta.group(cls=IgnoreRequiredWithHelp)
@click.pass_context
@click.option(
    "--catalog",
    "-c",
    show_envvar=True,
    help="Path to the tests catalog YAML file",
    type=click.Path(file_okay=True, dir_okay=False, exists=True, readable=True),
    required=True,
    callback=parse_catalog,
)
def nrfu(ctx: click.Context, catalog: List[Tuple[Callable[..., TestResult], Dict[Any, Any]]]) -> None:
    """Run NRFU against inventory devices"""
    ctx.obj["catalog"] = catalog


@anta.group("exec")
def _exec() -> None:
    """Execute commands to inventory devices"""


@anta.group("get")
def _get() -> None:
    """Get data from/to ANTA"""


@anta.group("debug")
def _debug() -> None:
    """Debug commands for building ANTA"""


# Load group commands
# Prefixing with `_` for avoiding the confusion when importing anta.cli.debug.commands as otherwise the debug group has
# a commands attribute.
_exec.add_command(exec_commands.clear_counters)
_exec.add_command(exec_commands.snapshot)
_exec.add_command(exec_commands.collect_tech_support)


_get.add_command(get_commands.from_cvp)
_get.add_command(get_commands.from_ansible)
_get.add_command(get_commands.inventory)
_get.add_command(get_commands.tags)

_debug.add_command(debug_commands.run_cmd)
_debug.add_command(debug_commands.run_template)

nrfu.add_command(check_commands.table)
nrfu.add_command(check_commands.json)
nrfu.add_command(check_commands.text)
nrfu.add_command(check_commands.tpl_report)


# ANTA CLI Execution
def cli() -> None:
    """Entrypoint for pyproject.toml"""
    anta(obj={}, auto_envvar_prefix="ANTA")  # pragma: no cover


if __name__ == "__main__":
    cli()



#!/usr/bin/python
# coding: utf-8 -*-
"""
Utils functions to use with anta.cli module.
"""
from __future__ import annotations

import enum
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple

import click
from yaml import safe_load

import anta.loader
from anta.inventory import AntaInventory
from anta.result_manager.models import TestResult
from anta.tools.misc import anta_log_exception

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from click import Option

    from anta.result_manager import ResultManager


class ExitCode(enum.IntEnum):
    """
    Encodes the valid exit codes by anta
    inspired from pytest
    """

    # Tests passed.
    OK = 0
    #: Tests failed.
    TESTS_FAILED = 1
    # Test error
    TESTS_ERROR = 2
    # An internal error got in the way.
    INTERNAL_ERROR = 3
    #  pytest was misused.
    USAGE_ERROR = 4


def parse_inventory(ctx: click.Context, path: Path) -> AntaInventory:
    """
    Helper function parse an ANTA inventory YAML file
    """
    if ctx.obj.get("_anta_help"):
        # Currently looking for help for a subcommand so no
        # need to parse the Inventory, return an empty one
        return AntaInventory()
    try:
        inventory = AntaInventory.parse(
            inventory_file=str(path),
            username=ctx.params["username"],
            password=ctx.params["password"],
            enable=ctx.params["enable"],
            enable_password=ctx.params["enable_password"],
            timeout=ctx.params["timeout"],
            insecure=ctx.params["insecure"],
        )
    except Exception as e:  # pylint: disable=broad-exception-caught
        message = f"Unable to parse ANTA Inventory file '{path}'"
        anta_log_exception(e, message, logger)
        ctx.fail(message)
    return inventory


def parse_tags(ctx: click.Context, param: Option, value: str) -> Optional[List[str]]:
    # pylint: disable=unused-argument
    """
    Click option callback to parse an ANTA inventory tags
    """
    if value is not None:
        return value.split(",") if "," in value else [value]
    return None


def requires_enable(ctx: click.Context, param: Option, value: Optional[str]) -> Optional[str]:
    # pylint: disable=unused-argument
    """
    Click option callback to ensure that enable is True when the option is set
    """
    if value is not None and ctx.params.get("enable") is not True:
        raise click.BadParameter(f"'{param.opts[0]}' requires '--enable' (or setting associated env variable)")
    return value


def parse_catalog(ctx: click.Context, param: Option, value: str) -> List[Tuple[Callable[..., TestResult], Dict[Any, Any]]]:
    # pylint: disable=unused-argument
    """
    Click option callback to parse an ANTA tests catalog YAML file
    """
    if ctx.obj.get("_anta_help"):
        # Currently looking for help for a subcommand so no
        # need to parse the Catalog - return an empty list
        return []
    try:
        with open(value, "r", encoding="UTF-8") as file:
            data = safe_load(file)
    # TODO catch proper exception
    # pylint: disable-next=broad-exception-caught
    except Exception as e:
        message = f"Unable to parse ANTA Tests Catalog file '{value}'"
        anta_log_exception(e, message, logger)
        ctx.fail(message)

    return anta.loader.parse_catalog(data)


def setup_logging(ctx: click.Context, param: Option, value: str) -> str:
    # pylint: disable=unused-argument
    """
    Click option callback to set ANTA logging level
    """
    try:
        anta.loader.setup_logging(value)
    except Exception as e:  # pylint: disable=broad-exception-caught
        message = f"Unable to set ANTA logging level '{value}'"
        anta_log_exception(e, message, logger)
        ctx.fail(message)

    return value


def return_code(result_manager: ResultManager, ignore_error: bool, ignore_status: bool) -> int:
    """
    Args:
        result_manager (ResultManager)
        ignore_error (bool): Ignore error status
        ignore_status (bool): Ignore status completely and always return 0

    Returns:
        exit_code (int):
          * 0 if ignore_status is True or status is in ["unset", "skipped", "success"]
          * 1 if status is "failure"
          * 2 if status is "error"
    """

    if ignore_status:
        return 0

    # If ignore_error is True then status can never be "error"
    status = result_manager.get_status(ignore_error=ignore_error)

    if status in {"unset", "skipped", "success"}:
        return ExitCode.OK
    if status == "failure":
        return ExitCode.TESTS_FAILED
    if status == "error":
        return ExitCode.TESTS_ERROR

    logger.error("Please gather logs and open an issue on Github.")
    raise ValueError(f"Unknown status returned by the ResultManager: {status}. Please gather logs and open an issue on Github.")


class IgnoreRequiredWithHelp(click.Group):
    """
    https://stackoverflow.com/questions/55818737/python-click-application-required-parameters-have-precedence-over-sub-command-he
    Solution to allow help without required options on subcommand

    This is not planned to be fixed in click as per: https://github.com/pallets/click/issues/295#issuecomment-708129734
    """

    def parse_args(self, ctx: click.Context, args: List[str]) -> List[str]:
        """
        Ignore MissingParameter exception when parsing arguments if `--help`
        is present for a subcommand
        """
        # Adding a flag for potential callbacks
        ctx.ensure_object(dict)
        if "--help" in args:
            ctx.obj["_anta_help"] = True

        try:
            return super().parse_args(ctx, args)
        except click.MissingParameter:
            if "--help" not in args:
                raise

            # remove the required params so that help can display
            for param in self.params:
                param.required = False

            return super().parse_args(ctx, args)



"""
Test functions related to Multi-chassis Link Aggregation (MLAG)
"""
from typing import Optional

from anta.models import AntaCommand, AntaTest
from anta.tools.get_value import get_value


class VerifyMlagStatus(AntaTest):
    """
    This test verifies the health status of the MLAG configuration.

    Expected Results:
        * success: The test will pass if the MLAG state is 'active', negotiation status is 'connected',
                   peer-link status and local interface status are 'up'.
        * failure: The test will fail if the MLAG state is not 'active', negotiation status is not 'connected',
                   peer-link status or local interface status are not 'up'.
        * skipped: The test will be skipped if MLAG is 'disabled'.
    """

    name = "VerifyMlagStatus"
    description = "This test verifies the health status of the MLAG configuration."
    categories = ["mlag"]
    commands = [AntaCommand(command="show mlag", ofmt="json")]

    @AntaTest.anta_test
    def test(self) -> None:
        """
        Run VerifyMlagStatus validation
        """

        command_output = self.instance_commands[0].json_output

        if command_output["state"] == "disabled":
            self.result.is_skipped("MLAG is disabled")
            return

        keys_to_verify = ["state", "negStatus", "localIntfStatus", "peerLinkStatus"]
        verified_output = {key: get_value(command_output, key) for key in keys_to_verify}

        if (
            verified_output["state"] == "active"
            and verified_output["negStatus"] == "connected"
            and verified_output["localIntfStatus"] == "up"
            and verified_output["peerLinkStatus"] == "up"
        ):
            self.result.is_success()
        else:
            self.result.is_failure(f"MLAG status is not OK: {verified_output}")


class VerifyMlagInterfaces(AntaTest):
    """
    This test verifies there are no inactive or active-partial MLAG ports.

    Expected Results:
        * success: The test will pass if there are NO inactive or active-partial MLAG ports.
        * failure: The test will fail if there are inactive or active-partial MLAG ports.
        * skipped: The test will be skipped if MLAG is 'disabled'.
    """

    name = "VerifyMlagInterfaces"
    description = "This test verifies there are no inactive or active-partial MLAG ports."
    categories = ["mlag"]
    commands = [AntaCommand(command="show mlag", ofmt="json")]

    @AntaTest.anta_test
    def test(self) -> None:
        """
        Run VerifyMlagInterfaces validation
        """

        command_output = self.instance_commands[0].json_output

        if command_output["state"] == "disabled":
            self.result.is_skipped("MLAG is disabled")
            return

        if command_output["mlagPorts"]["Inactive"] == 0 and command_output["mlagPorts"]["Active-partial"] == 0:
            self.result.is_success()
        else:
            self.result.is_failure(f"MLAG status is not OK: {command_output['mlagPorts']}")


class VerifyMlagConfigSanity(AntaTest):
    """
    This test verifies there are no MLAG config-sanity inconsistencies.

    Expected Results:
        * success: The test will pass if there are NO MLAG config-sanity inconsistencies.
        * failure: The test will fail if there are MLAG config-sanity inconsistencies.
        * skipped: The test will be skipped if MLAG is 'disabled'.
        * error: The test will give an error if 'mlagActive' is not found in the JSON response.
    """

    name = "VerifyMlagConfigSanity"
    description = "This test verifies there are no MLAG config-sanity inconsistencies."
    categories = ["mlag"]
    commands = [AntaCommand(command="show mlag config-sanity", ofmt="json")]

    @AntaTest.anta_test
    def test(self) -> None:
        """
        Run VerifyMlagConfigSanity validation
        """

        command_output = self.instance_commands[0].json_output

        if (mlag_status := get_value(command_output, "mlagActive")) is None:
            self.result.is_error("Incorrect JSON response - 'mlagActive' state was not found")
            return

        if mlag_status is False:
            self.result.is_skipped("MLAG is disabled")
            return

        keys_to_verify = ["globalConfiguration", "interfaceConfiguration"]
        verified_output = {key: get_value(command_output, key) for key in keys_to_verify}

        if not any(verified_output.values()):
            self.result.is_success()
        else:
            self.result.is_failure(f"MLAG config-sanity returned inconsistencies: {verified_output}")


class VerifyMlagReloadDelay(AntaTest):
    """
    This test verifies the reload-delay parameters of the MLAG configuration.

    Expected Results:
        * success: The test will pass if the reload-delay parameters are configured properly.
        * failure: The test will fail if the reload-delay parameters are NOT configured properly.
        * skipped: The test will be skipped if the reload-delay parameters are NOT provided or if MLAG is 'disabled'.
    """

    name = "VerifyMlagReloadDelay"
    description = "This test verifies the reload-delay parameters of the MLAG configuration."
    categories = ["mlag"]
    commands = [AntaCommand(command="show mlag", ofmt="json")]

    @AntaTest.anta_test
    def test(self, reload_delay: Optional[int] = None, reload_delay_non_mlag: Optional[int] = None) -> None:
        """
        Run VerifyMlagReloadDelay validation

        Args:
            reload_delay: Delay (seconds) after reboot until non peer-link ports that are part of an MLAG are enabled.
            reload_delay_non_mlag: Delay (seconds) after reboot until ports that are not part of an MLAG are enabled.
        """

        if not reload_delay or not reload_delay_non_mlag:
            self.result.is_skipped(f"{self.__class__.name} did not run because reload_delay or reload_delay_non_mlag were not supplied")
            return

        command_output = self.instance_commands[0].json_output

        if command_output["state"] == "disabled":
            self.result.is_skipped("MLAG is disabled")
            return

        keys_to_verify = ["reloadDelay", "reloadDelayNonMlag"]
        verified_output = {key: get_value(command_output, key) for key in keys_to_verify}

        if verified_output["reloadDelay"] == reload_delay and verified_output["reloadDelayNonMlag"] == reload_delay_non_mlag:
            self.result.is_success()

        else:
            self.result.is_failure(f"The reload-delay parameters are not configured properly: {verified_output}")


class VerifyMlagDualPrimary(AntaTest):
    """
    This test verifies the dual-primary detection and its parameters of the MLAG configuration.

    Expected Results:
        * success: The test will pass if the dual-primary detection is enabled and its parameters are configured properly.
        * failure: The test will fail if the dual-primary detection is NOT enabled or its parameters are NOT configured properly.
        * skipped: The test will be skipped if the dual-primary parameters are NOT provided or if MLAG is 'disabled'.
    """

    name = "VerifyMlagDualPrimary"
    description = "This test verifies the dual-primary detection and its parameters of the MLAG configuration."
    categories = ["mlag"]
    commands = [AntaCommand(command="show mlag detail", ofmt="json")]

    @AntaTest.anta_test
    def test(
        self, detection_delay: Optional[int] = None, errdisabled: bool = False, recovery_delay: Optional[int] = None, recovery_delay_non_mlag: Optional[int] = None
    ) -> None:
        """
        Run VerifyMlagDualPrimary validation

        Args:
            detection_delay: Delay detection for <N> seconds.
            errdisabled: Errdisabled all interfaces when dual-primary is detected. Defaults to False.
            recovery_delay: Delay (seconds) after dual-primary detection resolves until non peer-link ports that are part of an MLAG are enabled.
            recovery_delay_non_mlag: Delay (seconds) after dual-primary detection resolves until ports that are not part of an MLAG are enabled.
        """

        if detection_delay is None or errdisabled is None or recovery_delay is None or recovery_delay_non_mlag is None:
            self.result.is_skipped(
                f"{self.__class__.name} did not run because detection_delay, errdisabled, recovery_delay or recovery_delay_non_mlag were not supplied"
            )
            return

        errdisabled_action = "errdisableAllInterfaces" if errdisabled else "none"

        command_output = self.instance_commands[0].json_output

        if command_output["state"] == "disabled":
            self.result.is_skipped("MLAG is disabled")
            return

        if command_output["dualPrimaryDetectionState"] == "disabled":
            self.result.is_failure("Dual-primary detection is disabled")
            return

        keys_to_verify = ["detail.dualPrimaryDetectionDelay", "detail.dualPrimaryAction", "dualPrimaryMlagRecoveryDelay", "dualPrimaryNonMlagRecoveryDelay"]
        verified_output = {key: get_value(command_output, key) for key in keys_to_verify}

        if (
            verified_output["detail.dualPrimaryDetectionDelay"] == detection_delay
            and verified_output["detail.dualPrimaryAction"] == errdisabled_action
            and verified_output["dualPrimaryMlagRecoveryDelay"] == recovery_delay
            and verified_output["dualPrimaryNonMlagRecoveryDelay"] == recovery_delay_non_mlag
        ):
            self.result.is_success()

        else:
            self.result.is_failure(f"The dual-primary parameters are not configured properly: {verified_output}")



"""
Test functions related to ASIC profiles
"""

from typing import Optional

from anta.decorators import skip_on_platforms
from anta.models import AntaCommand, AntaTest


class VerifyUnifiedForwardingTableMode(AntaTest):
    """
    Verifies the device is using the expected Unified Forwarding Table mode.
    """

    name = "VerifyUnifiedForwardingTableMode"
    description = ""
    categories = ["profiles"]
    commands = [AntaCommand(command="show platform trident forwarding-table partition", ofmt="json")]

    @skip_on_platforms(["cEOSLab", "vEOS-lab"])
    @AntaTest.anta_test
    def test(self, mode: Optional[str] = None) -> None:
        """
        Run VerifyUnifiedForwardingTableMode validation

        Args:
            mode: Expected UFT mode.
        """
        if not mode:
            self.result.is_skipped("VerifyUnifiedForwardingTableMode was not run as no mode was given")
            return

        command_output = self.instance_commands[0].json_output
        if command_output["uftMode"] == mode:
            self.result.is_success()
        else:
            self.result.is_failure(f"Device is not running correct UFT mode (expected: {mode} / running: {command_output['uftMode']})")


class VerifyTcamProfile(AntaTest):
    """
    Verifies the device is using the configured TCAM profile.
    """

    name = "VerifyTcamProfile"
    description = "Verify that the assigned TCAM profile is actually running on the device"
    categories = ["profiles"]
    commands = [AntaCommand(command="show hardware tcam profile", ofmt="json")]

    @skip_on_platforms(["cEOSLab", "vEOS-lab"])
    @AntaTest.anta_test
    def test(self, profile: Optional[str] = None) -> None:
        """
        Run VerifyTcamProfile validation

        Args:
            profile: Expected TCAM profile.
        """
        if not profile:
            self.result.is_skipped("VerifyTcamProfile was not run as no profile was given")
            return

        command_output = self.instance_commands[0].json_output
        if command_output["pmfProfiles"]["FixedSystem"]["status"] == command_output["pmfProfiles"]["FixedSystem"]["config"] == profile:
            self.result.is_success()
        else:
            self.result.is_failure(f"Incorrect profile running on device: {command_output['pmfProfiles']['FixedSystem']['status']}")



"""
Test functions related to the EOS various security settings
"""
from __future__ import annotations

from typing import Optional

from anta.models import AntaCommand, AntaTest


class VerifySSHStatus(AntaTest):
    """
    Verifies if the SSHD agent is disabled in the default VRF.

    Expected Results:
        * success: The test will pass if the SSHD agent is disabled in the default VRF.
        * failure: The test will fail if the SSHD agent is NOT disabled in the default VRF.
    """

    name = "VerifySSHStatus"
    description = "Verifies if the SSHD agent is disabled in the default VRF."
    categories = ["security"]
    commands = [AntaCommand(command="show management ssh", ofmt="text")]

    @AntaTest.anta_test
    def test(self) -> None:
        """
        Run VerifySSHStatus validation.
        """

        command_output = self.instance_commands[0].text_output

        line = [line for line in command_output.split("\n") if line.startswith("SSHD status")][0]
        status = line.split("is ")[1]

        if status == "disabled":
            self.result.is_success()
        else:
            self.result.is_failure(line)


class VerifySSHIPv4Acl(AntaTest):
    """
    Verifies if the SSHD agent has the right number IPv4 ACL(s) configured for a specified VRF.

    Expected results:
        * success: The test will pass if the SSHD agent has the provided number of IPv4 ACL(s) in the specified VRF.
        * failure: The test will fail if the SSHD agent has not the right number of IPv4 ACL(s) in the specified VRF.
        * skipped: The test will be skipped if the number of IPv4 ACL(s) or VRF parameter is not provided.
    """

    name = "VerifySSHIPv4Acl"
    description = "Verifies if the SSHD agent has IPv4 ACL(s) configured."
    categories = ["security"]
    commands = [AntaCommand(command="show management ssh ip access-list summary")]

    @AntaTest.anta_test
    def test(self, number: Optional[int] = None, vrf: str = "default") -> None:
        """
        Run VerifySSHIPv4Acl validation.

        Args:
            number: The number of expected IPv4 ACL(s).
            vrf: The name of the VRF in which to check for the SSHD agent. Defaults to 'default'.
        """
        if not number or not vrf:
            self.result.is_skipped(f"{self.__class__.name} did not run because number or vrf was not supplied")
            return

        command_output = self.instance_commands[0].json_output

        ipv4_acl_list = command_output["ipAclList"]["aclList"]
        ipv4_acl_number = len(ipv4_acl_list)
        not_configured_acl_list = []

        if ipv4_acl_number != number:
            self.result.is_failure(f"Expected {number} SSH IPv4 ACL(s) in vrf {vrf} but got {ipv4_acl_number}")
            return

        for ipv4_acl in ipv4_acl_list:
            if vrf not in ipv4_acl["configuredVrfs"] or vrf not in ipv4_acl["activeVrfs"]:
                not_configured_acl_list.append(ipv4_acl["name"])

        if not_configured_acl_list:
            self.result.is_failure(f"SSH IPv4 ACL(s) not configured or active in vrf {vrf}: {not_configured_acl_list}")
        else:
            self.result.is_success()


class VerifySSHIPv6Acl(AntaTest):
    """
    Verifies if the SSHD agent has the right number IPv6 ACL(s) configured for a specified VRF.

    Expected results:
        * success: The test will pass if the SSHD agent has the provided number of IPv6 ACL(s) in the specified VRF.
        * failure: The test will fail if the SSHD agent has not the right number of IPv6 ACL(s) in the specified VRF.
        * skipped: The test will be skipped if the number of IPv6 ACL(s) or VRF parameter is not provided.
    """

    name = "VerifySSHIPv6Acl"
    description = "Verifies if the SSHD agent has IPv6 ACL(s) configured."
    categories = ["security"]
    commands = [AntaCommand(command="show management ssh ipv6 access-list summary")]

    @AntaTest.anta_test
    def test(self, number: Optional[int] = None, vrf: str = "default") -> None:
        """
        Run VerifySSHIPv6Acl validation.

        Args:
            number: The number of expected IPv6 ACL(s).
            vrf: The name of the VRF in which to check for the SSHD agent. Defaults to 'default'.
        """
        if not number or not vrf:
            self.result.is_skipped(f"{self.__class__.name} did not run because number or vrf was not supplied")
            return

        command_output = self.instance_commands[0].json_output

        ipv6_acl_list = command_output["ipv6AclList"]["aclList"]
        ipv6_acl_number = len(ipv6_acl_list)
        not_configured_acl_list = []

        if ipv6_acl_number != number:
            self.result.is_failure(f"Expected {number} SSH IPv6 ACL(s) in vrf {vrf} but got {ipv6_acl_number}")
            return

        for ipv6_acl in ipv6_acl_list:
            if vrf not in ipv6_acl["configuredVrfs"] or vrf not in ipv6_acl["activeVrfs"]:
                not_configured_acl_list.append(ipv6_acl["name"])

        if not_configured_acl_list:
            self.result.is_failure(f"SSH IPv6 ACL(s) not configured or active in vrf {vrf}: {not_configured_acl_list}")
        else:
            self.result.is_success()


class VerifyTelnetStatus(AntaTest):
    """
    Verifies if Telnet is disabled in the default VRF.

    Expected Results:
        * success: The test will pass if Telnet is disabled in the default VRF.
        * failure: The test will fail if Telnet is NOT disabled in the default VRF.
    """

    name = "VerifyTelnetStatus"
    description = "Verifies if Telnet is disabled in the default VRF."
    categories = ["security"]
    commands = [AntaCommand(command="show management telnet")]

    @AntaTest.anta_test
    def test(self) -> None:
        """
        Run VerifyTelnetStatus validation.
        """

        command_output = self.instance_commands[0].json_output

        if command_output["serverState"] == "disabled":
            self.result.is_success()
        else:
            self.result.is_failure("Telnet status for Default VRF is enabled")


class VerifyAPIHttpStatus(AntaTest):
    """
    Verifies if eAPI HTTP server is disabled globally.

    Expected Results:
        * success: The test will pass if eAPI HTTP server is disabled globally.
        * failure: The test will fail if eAPI HTTP server is NOT disabled globally.
    """

    name = "VerifyAPIHttpStatus"
    description = "Verifies if eAPI HTTP server is disabled globally."
    categories = ["security"]
    commands = [AntaCommand(command="show management api http-commands")]

    @AntaTest.anta_test
    def test(self) -> None:
        """
        Run VerifyAPIHTTPStatus validation.
        """

        command_output = self.instance_commands[0].json_output

        if command_output["enabled"] and not command_output["httpServer"]["running"]:
            self.result.is_success()
        else:
            self.result.is_failure("eAPI HTTP server is enabled globally")


class VerifyAPIHttpsSSL(AntaTest):
    """
    Verifies if eAPI HTTPS server SSL profile is configured and valid.

    Expected results:
        * success: The test will pass if the eAPI HTTPS server SSL profile is configured and valid.
        * failure: The test will fail if the eAPI HTTPS server SSL profile is NOT configured, misconfigured or invalid.
        * skipped: The test will be skipped if the SSL profile is not provided.
    """

    name = "VerifyAPIHttpsSSL"
    description = "Verifies if eAPI HTTPS server SSL profile is configured and valid."
    categories = ["security"]
    commands = [AntaCommand(command="show management api http-commands")]

    @AntaTest.anta_test
    def test(self, profile: Optional[str] = None) -> None:
        """
        Run VerifyAPIHttpsSSL validation.

        Args:
            profile: SSL profile to verify.
        """
        if not profile:
            self.result.is_skipped(f"{self.__class__.name} did not run because profile was not supplied")
            return

        command_output = self.instance_commands[0].json_output

        try:
            if command_output["sslProfile"]["name"] == profile and command_output["sslProfile"]["state"] == "valid":
                self.result.is_success()
            else:
                self.result.is_failure(f"eAPI HTTPS server SSL profile ({profile}) is misconfigured or invalid")

        except KeyError:
            self.result.is_failure(f"eAPI HTTPS server SSL profile ({profile}) is not configured")


class VerifyAPIIPv4Acl(AntaTest):
    """
    Verifies if eAPI has the right number IPv4 ACL(s) configured for a specified VRF.

    Expected results:
        * success: The test will pass if eAPI has the provided number of IPv4 ACL(s) in the specified VRF.
        * failure: The test will fail if eAPI has not the right number of IPv4 ACL(s) in the specified VRF.
        * skipped: The test will be skipped if the number of IPv4 ACL(s) or VRF parameter is not provided.
    """

    name = "VerifyAPIIPv4Acl"
    description = "Verifies if eAPI has the right number IPv4 ACL(s) configured for a specified VRF."
    categories = ["security"]
    commands = [AntaCommand(command="show management api http-commands ip access-list summary")]

    @AntaTest.anta_test
    def test(self, number: Optional[int] = None, vrf: str = "default") -> None:
        """
        Run VerifyAPIIPv4Acl validation.

        Args:
            number: The number of expected IPv4 ACL(s).
            vrf: The name of the VRF in which to check for eAPI. Defaults to 'default'.
        """
        if not number or not vrf:
            self.result.is_skipped(f"{self.__class__.name} did not run because number or vrf was not supplied")
            return

        command_output = self.instance_commands[0].json_output

        ipv4_acl_list = command_output["ipAclList"]["aclList"]
        ipv4_acl_number = len(ipv4_acl_list)
        not_configured_acl_list = []

        if ipv4_acl_number != number:
            self.result.is_failure(f"Expected {number} eAPI IPv4 ACL(s) in vrf {vrf} but got {ipv4_acl_number}")
            return

        for ipv4_acl in ipv4_acl_list:
            if vrf not in ipv4_acl["configuredVrfs"] or vrf not in ipv4_acl["activeVrfs"]:
                not_configured_acl_list.append(ipv4_acl["name"])

        if not_configured_acl_list:
            self.result.is_failure(f"eAPI IPv4 ACL(s) not configured or active in vrf {vrf}: {not_configured_acl_list}")
        else:
            self.result.is_success()


class VerifyAPIIPv6Acl(AntaTest):
    """
    Verifies if eAPI has the right number IPv6 ACL(s) configured for a specified VRF.

    Expected results:
        * success: The test will pass if eAPI has the provided number of IPv6 ACL(s) in the specified VRF.
        * failure: The test will fail if eAPI has not the right number of IPv6 ACL(s) in the specified VRF.
        * skipped: The test will be skipped if the number of IPv6 ACL(s) or VRF parameter is not provided.
    """

    name = "VerifyAPIIPv6Acl"
    description = "Verifies if eAPI has the right number IPv6 ACL(s) configured for a specified VRF."
    categories = ["security"]
    commands = [AntaCommand(command="show management api http-commands ipv6 access-list summary")]

    @AntaTest.anta_test
    def test(self, number: Optional[int] = None, vrf: str = "default") -> None:
        """
        Run VerifyAPIIPv6Acl validation.

        Args:
            number: The number of expected IPv6 ACL(s).
            vrf: The name of the VRF in which to check for eAPI. Defaults to 'default'.
        """
        if not number or not vrf:
            self.result.is_skipped(f"{self.__class__.name} did not run because number or vrf was not supplied")
            return

        command_output = self.instance_commands[0].json_output

        ipv6_acl_list = command_output["ipv6AclList"]["aclList"]
        ipv6_acl_number = len(ipv6_acl_list)
        not_configured_acl_list = []

        if ipv6_acl_number != number:
            self.result.is_failure(f"Expected {number} eAPI IPv6 ACL(s) in vrf {vrf} but got {ipv6_acl_number}")
            return

        for ipv6_acl in ipv6_acl_list:
            if vrf not in ipv6_acl["configuredVrfs"] or vrf not in ipv6_acl["activeVrfs"]:
                not_configured_acl_list.append(ipv6_acl["name"])

        if not_configured_acl_list:
            self.result.is_failure(f"eAPI IPv6 ACL(s) not configured or active in vrf {vrf}: {not_configured_acl_list}")
        else:
            self.result.is_success()



"""
Models to define a TestStructure
"""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from copy import deepcopy
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Coroutine, Dict, List, Literal, Optional, TypeVar, Union

from pydantic import BaseModel, ConfigDict, conint
from rich.progress import Progress, TaskID

from anta.result_manager.models import TestResult
from anta.tools.misc import anta_log_exception, exc_to_str

if TYPE_CHECKING:
    from anta.device import AntaDevice

F = TypeVar("F", bound=Callable[..., Any])
DEFAULT_TAG = "all"

logger = logging.getLogger(__name__)

# TODO: Notes on eAPI version/revision
# eAPI models are revisioned, this means that if a model is modified in a non-backwards compatible way, then its revision will be bumped up
# (revisions are numbers, default value is 1).
# By default an eAPI request will return revision 1 of the model instance,
# this ensures that older management software will not suddenly stop working when a switch is upgraded.
# A "revision" applies to a particular CLI command whereas a "version" is global and is internally
# translated to a specific "revision" for each CLI command in the rpc.


class AntaTemplate(BaseModel):
    """Class to define a test command with its API version

    Attributes:
        template: Python f-string. Example: 'show vlan {vlan_id}'
        version: eAPI version - valid values are 1 or "latest" - default is "latest"
        revision: Revision of the command. Valid values are 1 to 99. Revision has precedence over version.
        ofmt: eAPI output - json or text - default is json
    """

    template: str
    version: Literal[1, "latest"] = "latest"
    revision: Optional[conint(ge=1, le=99)] = None  # type: ignore
    ofmt: Literal["json", "text"] = "json"

    def render(self, params: Dict[str, Any]) -> AntaCommand:
        """Render an AntaCommand from an AntaTemplate instance.
        Keep the parameters used in the AntaTemplate instance.

         Args:
             params: dictionary of variables with string values to render the Python f-string

         Returns:
             AntaCommand: The rendered AntaCommand.
                          This AntaCommand instance have a template attribute that references this
                          AntaTemplate instance.
        """
        return AntaCommand(command=self.template.format(**params), ofmt=self.ofmt, version=self.version, revision=self.revision, template=self, params=params)


class AntaCommand(BaseModel):
    """Class to define a test command with its API version

    Attributes:
        command: Device command
        version: eAPI version - valid values are 1 or "latest" - default is "latest"
        revision: Revision of the command. Valid values are 1 to 99. Revision has precedence over version.
        ofmt: eAPI output - json or text - default is json
        template: AntaTemplate object used to render this command
        params: dictionary of variables with string values to render the template
        failed: If the command execution fails, the Exception object is stored in this field
    """

    # This is required if we want to keep an Exception object in the failed field
    model_config = ConfigDict(arbitrary_types_allowed=True)

    command: str
    version: Literal[1, "latest"] = "latest"
    revision: Optional[conint(ge=1, le=99)] = None  # type: ignore
    ofmt: Literal["json", "text"] = "json"
    output: Optional[Union[Dict[str, Any], str]] = None
    template: Optional[AntaTemplate] = None
    failed: Optional[Exception] = None
    params: Optional[Dict[str, Any]] = None

    @property
    def json_output(self) -> Dict[str, Any]:
        """Get the command output as JSON"""
        if self.output is None:
            raise RuntimeError(f"There is no output for command {self.command}")
        if self.ofmt != "json" or not isinstance(self.output, dict):
            raise RuntimeError(f"Output of command {self.command} is invalid")
        return dict(self.output)

    @property
    def text_output(self) -> str:
        """Get the command output as a string"""
        if self.output is None:
            raise RuntimeError(f"There is no output for command {self.command}")
        if self.ofmt != "text" or not isinstance(self.output, str):
            raise RuntimeError(f"Output of command {self.command} is invalid")
        return str(self.output)

    @property
    def collected(self) -> bool:
        """Return True if the command has been collected"""
        return self.output is not None and self.failed is None


class AntaTestFilter(ABC):
    """Class to define a test Filter"""

    # pylint: disable=too-few-public-methods

    @abstractmethod
    def should_skip(
        self,
        device: AntaDevice,
        result: TestResult,
        *args: list[Any],
        **kwagrs: dict[str, Any],
    ) -> bool:
        """
        Sets the TestResult status to skip with the appropriate skip message

        Returns:
            bool: True if the test should be skipped, False otherwise
        """


class AntaTest(ABC):
    """Abstract class defining a test for Anta

    The goal of this class is to handle the heavy lifting and make
    writing a test as simple as possible.

    TODO - complete doctstring with example
    """

    # pylint: disable=too-many-instance-attributes

    # Mandatory class attributes
    # TODO - find a way to tell mypy these are mandatory for child classes - maybe Protocol
    name: ClassVar[str]
    description: ClassVar[str]
    categories: ClassVar[list[str]]
    # Or any child type
    commands: ClassVar[list[AntaCommand]]
    # TODO - today we support only one template per Test
    template: ClassVar[AntaTemplate]
    progress: Optional[Progress] = None
    nrfu_task: Optional[TaskID] = None

    # Optional class attributes
    test_filters: ClassVar[list[AntaTestFilter]]

    def __init__(
        self,
        device: AntaDevice,
        template_params: list[dict[str, Any]] | None = None,
        result_description: str | None = None,
        result_categories: list[str] | None = None,
        result_custom_field: str | None = None,
        # TODO document very well the order of eos_data
        eos_data: list[dict[Any, Any] | str] | None = None,
        labels: list[str] | None = None,
    ):
        """
        AntaTest Constructor

        Doc to be completed

        Arguments:
            result_custom_field (str): a free string that is included in the TestResult object
        """
        # Accept 6 input arguments
        # pylint: disable=R0913
        self.logger: logging.Logger = logging.getLogger(f"{self.__module__}.{self.__class__.__name__}")
        self.device: AntaDevice = device
        self.result: TestResult = TestResult(
            name=device.name,
            test=self.name,
            categories=result_categories or self.categories,
            description=result_description or self.description,
            custom_field=result_custom_field,
        )
        self.labels: List[str] = labels or []
        self.instance_commands: List[AntaCommand] = []

        # TODO - check optimization for deepcopy
        # Generating instance_commands from list of commands and template
        if hasattr(self.__class__, "commands") and (cmds := self.__class__.commands) is not None:
            self.instance_commands.extend(deepcopy(cmds))
        if hasattr(self.__class__, "template") and (tpl := self.__class__.template) is not None:
            if template_params is None:
                self.result.is_error("Command has template but no params were given")
                return
            self.template_params = template_params
            for param in template_params:
                try:
                    self.instance_commands.append(tpl.render(param))
                except KeyError:
                    self.result.is_error(f"Cannot render template '{tpl.template}': wrong parameters")
                    return

        if eos_data is not None:
            self.logger.debug("Test initialized with input data")
            self.save_commands_data(eos_data)

    def save_commands_data(self, eos_data: list[dict[Any, Any] | str]) -> None:
        """Called at init or at test execution time"""
        if len(eos_data) != len(self.instance_commands):
            self.result.is_error("Test initialization error: Trying to save more data than there are commands for the test")
            return
        for index, data in enumerate(eos_data or []):
            self.instance_commands[index].output = data

    def all_data_collected(self) -> bool:
        """returns True if output is populated for every command"""
        return all(command.collected for command in self.instance_commands)

    def get_failed_commands(self) -> List[AntaCommand]:
        """returns a list of all the commands that have a populated failed field"""
        return [command for command in self.instance_commands if command.failed is not None]

    def __init_subclass__(cls) -> None:
        """
        Verify that the mandatory class attributes are defined
        """
        mandatory_attributes = ["name", "description", "categories"]
        for attr in mandatory_attributes:
            if not hasattr(cls, attr):
                raise NotImplementedError(f"Class {cls} is missing required class attribute {attr}")
        # Check that either commands or template exist
        if not (hasattr(cls, "commands") or hasattr(cls, "template")):
            raise NotImplementedError(f"Class {cls} is missing required either commands or template attribute")

    async def collect(self) -> None:
        """
        Method used to collect outputs of all commands of this test class from the device of this test instance.
        """
        try:
            await self.device.collect_commands(self.instance_commands)
        except Exception as e:  # pylint: disable=broad-exception-caught
            message = f"Exception raised while collecting commands for test {self.name} (on device {self.device.name})"
            anta_log_exception(e, message, self.logger)
            self.result.is_error(exc_to_str(e))

    @staticmethod
    def anta_test(function: F) -> Callable[..., Coroutine[Any, Any, TestResult]]:
        """
        Decorator for anta_test that handles injecting test data if given and collecting it using asyncio if missing
        """

        @wraps(function)
        async def wrapper(
            self: AntaTest,
            eos_data: list[dict[Any, Any] | str] | None = None,
            **kwargs: Any,
        ) -> TestResult:
            """
            Wraps the test function and implement (in this order):
            1. Instantiate the command outputs if `eos_data` is provided
            2. Collect missing command outputs from the device
            3. Run the test function
            4. Catches and set the result if the test function raises an exception

            Returns:
                TestResult: self.result, populated with the correct exit status
            """
            if self.result.result != "unset":
                return self.result

            # TODO maybe_skip decorators

            # Data
            if eos_data is not None:
                self.save_commands_data(eos_data)
                self.logger.debug(f"Test {self.name} initialized with input data {eos_data}")

            # If some data is missing, try to collect
            if not self.all_data_collected():
                await self.collect()
                if self.result.result != "unset":
                    return self.result

            try:
                if cmds := self.get_failed_commands():
                    self.result.is_error(
                        "\n".join([f"{cmd.command} has failed: {exc_to_str(cmd.failed)}" if cmd.failed else f"{cmd.command} has failed" for cmd in cmds])
                    )
                    return self.result
                function(self, **kwargs)
            except Exception as e:  # pylint: disable=broad-exception-caught
                message = f"Exception raised for test {self.name} (on device {self.device.name})"
                anta_log_exception(e, message, self.logger)
                self.result.is_error(exc_to_str(e))

            AntaTest.update_progress()
            return self.result

        return wrapper

    @classmethod
    def update_progress(cls) -> None:
        """
        Update progress bar for all AntaTest objects if it exists
        """
        if cls.progress and (cls.nrfu_task is not None):
            cls.progress.update(cls.nrfu_task, advance=1)

    @abstractmethod
    def test(self) -> Coroutine[Any, Any, TestResult]:
        """
        This abstract method is the core of the test.
        It MUST set the correct status of self.result with the appropriate error messages

        it must be implemented as follow

        @AntaTest.anta_test
        def test(self) -> None:
           '''
           assert code
           '''
        """



#!/usr/bin/env python
# coding: utf-8 -*-
"""
Commands for Anta CLI to run nrfu commands.
"""

import asyncio
import logging
import pathlib
import sys
from typing import List, Optional

import click

from anta.cli.utils import parse_tags, return_code
from anta.models import AntaTest
from anta.result_manager import ResultManager
from anta.runner import main

from .utils import anta_progress_bar, print_jinja, print_json, print_settings, print_table, print_text

logger = logging.getLogger(__name__)


@click.command()
@click.pass_context
@click.option("--tags", help="List of tags using comma as separator: tag1,tag2,tag3", type=str, required=False, callback=parse_tags)
@click.option("--device", "-d", help="Show a summary for this device", type=str, required=False)
@click.option("--test", "-t", help="Show a summary for this test", type=str, required=False)
@click.option(
    "--group-by", default=None, type=click.Choice(["device", "test"], case_sensitive=False), help="Group result by test or host. default none", required=False
)
def table(ctx: click.Context, tags: Optional[List[str]], device: Optional[str], test: Optional[str], group_by: str) -> None:
    """ANTA command to check network states with table result"""
    print_settings(ctx)
    results = ResultManager()
    with anta_progress_bar() as AntaTest.progress:
        asyncio.run(main(results, ctx.obj["inventory"], ctx.obj["catalog"], tags=tags))

    print_table(results=results, device=device, group_by=group_by, test=test)

    # TODO make a util method to avoid repeating the same three line
    ignore_status = ctx.obj["ignore_status"]
    ignore_error = ctx.obj["ignore_error"]
    sys.exit(return_code(results, ignore_error, ignore_status))


@click.command()
@click.pass_context
@click.option("--tags", "-t", help="List of tags using comma as separator: tag1,tag2,tag3", type=str, required=False, callback=parse_tags)
@click.option(
    "--output",
    "-o",
    type=click.Path(file_okay=True, dir_okay=False, exists=False, writable=True, path_type=pathlib.Path),
    show_envvar=True,
    required=False,
    help="Path to save report as a file",
)
def json(ctx: click.Context, tags: Optional[List[str]], output: Optional[pathlib.Path]) -> None:
    """ANTA command to check network state with JSON result"""
    print_settings(ctx)
    results = ResultManager()
    with anta_progress_bar() as AntaTest.progress:
        asyncio.run(main(results, ctx.obj["inventory"], ctx.obj["catalog"], tags=tags))
    print_json(results=results, output=output)

    ignore_status = ctx.obj["ignore_status"]
    ignore_error = ctx.obj["ignore_error"]
    sys.exit(return_code(results, ignore_error, ignore_status))


@click.command()
@click.pass_context
@click.option("--tags", "-t", help="List of tags using comma as separator: tag1,tag2,tag3", type=str, required=False, callback=parse_tags)
@click.option("--search", "-s", help="Regular expression to search in both name and test", type=str, required=False)
@click.option("--skip-error", help="Hide tests in errors due to connectivity issue", default=False, is_flag=True, show_default=True, required=False)
def text(ctx: click.Context, tags: Optional[List[str]], search: Optional[str], skip_error: bool) -> None:
    """ANTA command to check network states with text result"""
    print_settings(ctx)
    results = ResultManager()
    with anta_progress_bar() as AntaTest.progress:
        asyncio.run(main(results, ctx.obj["inventory"], ctx.obj["catalog"], tags=tags))
    print_text(results=results, search=search, skip_error=skip_error)

    ignore_status = ctx.obj["ignore_status"]
    ignore_error = ctx.obj["ignore_error"]
    sys.exit(return_code(results, ignore_error, ignore_status))


@click.command()
@click.pass_context
@click.option(
    "--template",
    "-tpl",
    type=click.Path(file_okay=True, dir_okay=False, exists=True, readable=True, path_type=pathlib.Path),
    show_envvar=True,
    required=True,
    help="Path to the template to use for the report",
)
@click.option(
    "--output",
    "-o",
    type=click.Path(file_okay=True, dir_okay=False, exists=False, writable=True, path_type=pathlib.Path),
    show_envvar=True,
    required=False,
    help="Path to save report as a file",
)
@click.option("--tags", "-t", help="List of tags using comma as separator: tag1,tag2,tag3", type=str, required=False, callback=parse_tags)
def tpl_report(ctx: click.Context, tags: Optional[List[str]], template: pathlib.Path, output: Optional[pathlib.Path]) -> None:
    """ANTA command to check network state with templated report"""
    print_settings(ctx, template, output)
    results = ResultManager()
    with anta_progress_bar() as AntaTest.progress:
        asyncio.run(main(results, ctx.obj["inventory"], ctx.obj["catalog"], tags=tags))
    print_jinja(results=results, template=template, output=output)

    ignore_status = ctx.obj["ignore_status"]
    ignore_error = ctx.obj["ignore_error"]
    sys.exit(return_code(results, ignore_error, ignore_status))



#!/usr/bin/env python
# coding: utf-8 -*-
# pylint: disable = redefined-outer-name

"""
Commands for Anta CLI to run debug commands.
"""

import asyncio
import logging
from typing import List, Literal

import click
from click import Option

from anta.cli.console import console
from anta.device import AntaDevice
from anta.models import AntaCommand, AntaTemplate
from anta.tools.misc import anta_log_exception

logger = logging.getLogger(__name__)


# pylint: disable-next=inconsistent-return-statements
def get_device(ctx: click.Context, param: Option, value: str) -> List[str]:
    """
    Click option callback to get an AntaDevice instance from a string
    """
    # pylint: disable=unused-argument
    try:
        return ctx.obj["inventory"][value]
    except KeyError as e:
        message = f"Device {value} does not exist in Inventory"
        anta_log_exception(e, message, logger)
        ctx.fail(message)


@click.command()
@click.option("--command", "-c", type=str, required=True, help="Command to run")
@click.option("--ofmt", type=click.Choice(["json", "text"]), default="json", help="EOS eAPI format to use. can be text or json")
@click.option("--version", "-v", type=click.Choice(["1", "latest"]), default="latest", help="EOS eAPI version")
@click.option("--revision", "-r", type=int, help="eAPI command revision", required=False)
@click.option("--device", "-d", type=str, required=True, help="Device from inventory to use", callback=get_device)
def run_cmd(command: str, ofmt: Literal["json", "text"], version: Literal["1", "latest"], revision: int, device: AntaDevice) -> None:
    """Run arbitrary command to an ANTA device"""
    console.print(f"Run command [green]{command}[/green] on [red]{device.name}[/red]")
    # I do not assume the following line, but click make me do it
    v: Literal[1, "latest"] = version if version == "latest" else 1
    c = AntaCommand(command=command, ofmt=ofmt, version=v, revision=revision)
    asyncio.run(device.collect(c))
    if ofmt == "json":
        console.print(c.json_output)
    if ofmt == "text":
        console.print(c.text_output)


@click.command()
@click.option("--template", "-t", type=str, required=True, help="Command template to run. E.g. 'show vlan {vlan_id}'")
@click.option("--ofmt", type=click.Choice(["json", "text"]), default="json", help="EOS eAPI format to use. can be text or json")
@click.option("--version", "-v", type=click.Choice(["1", "latest"]), default="latest", help="EOS eAPI version")
@click.option("--revision", "-r", type=int, help="eAPI command revision", required=False)
@click.option("--device", "-d", type=str, required=True, help="Device from inventory to use", callback=get_device)
@click.argument("params", required=True, nargs=-1)
def run_template(template: str, params: List[str], ofmt: Literal["json", "text"], version: Literal["1", "latest"], revision: int, device: AntaDevice) -> None:
    # pylint: disable=too-many-arguments
    """Run arbitrary templated command to an ANTA device.

    Takes a list of arguments (keys followed by a value) to build a dictionary used as template parameters.
    Example:

    anta debug run-template -d leaf1a -t 'show vlan {vlan_id}' vlan_id 1
    """
    template_params = dict(zip(params[::2], params[1::2]))

    console.print(f"Run templated command [blue]'{template}'[/blue] with [orange]{template_params}[/orange] on [red]{device.name}[/red]")
    # I do not assume the following line, but click make me do it
    v: Literal[1, "latest"] = version if version == "latest" else 1
    t = AntaTemplate(template=template, ofmt=ofmt, version=v, revision=revision)
    c = t.render(template_params)
    asyncio.run(device.collect(c))
    if ofmt == "json":
        console.print(c.json_output)
    if ofmt == "text":
        console.print(c.text_output)



"""
ANTA Device Abstraction Module
"""
import asyncio
import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, Iterator, List, Literal, Optional, Tuple, Union

import asyncssh
from aioeapi import Device, EapiCommandError
from asyncssh import SSHClientConnection, SSHClientConnectionOptions
from httpx import ConnectError, HTTPError

from anta import __DEBUG__
from anta.models import DEFAULT_TAG, AntaCommand
from anta.tools.misc import anta_log_exception, exc_to_str

logger = logging.getLogger(__name__)


class AntaDevice(ABC):
    """
    Abstract class representing a device in ANTA.
    An implementation of this class needs must override the abstract coroutines `collect()` and
    `refresh()`.

    Attributes:
        name: Device name
        is_online: True if the device IP is reachable and a port can be open
        established: True if remote command execution succeeds
        hw_model: Hardware model of the device
        tags: List of tags for this device
    """

    def __init__(self, name: str, tags: Optional[List[str]] = None) -> None:
        """
        Constructor of AntaDevice

        Args:
            name: Device name
            tags: List of tags for this device
        """
        self.name: str = name
        self.hw_model: Optional[str] = None
        self.tags: List[str] = tags if tags is not None else []
        self.is_online: bool = False
        self.established: bool = False

        # Ensure tag 'all' is always set
        if DEFAULT_TAG not in self.tags:
            self.tags.append(DEFAULT_TAG)

    def __rich_repr__(self) -> Iterator[Tuple[str, Any]]:
        """
        Implements Rich Repr Protocol
        https://rich.readthedocs.io/en/stable/pretty.html#rich-repr-protocol
        """
        yield "name", self.name
        yield "tags", self.tags
        yield "hw_model", self.hw_model
        yield "is_online", self.is_online
        yield "established", self.established

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        """
        AntaDevice equality depends on the class implementation.
        """

    @abstractmethod
    async def collect(self, command: AntaCommand) -> None:
        """
        Collect device command output.
        This abstract coroutine can be used to implement any command collection method
        for a device in ANTA.

        The `collect()` implementation needs to populate the `output` attribute
        of the `AntaCommand` object passed as argument.

        If a failure occurs, the `collect()` implementation is expected to catch the
        exception and implement proper logging, the `output` attribute of the
        `AntaCommand` object passed as argument would be `None` in this case.

        Args:
            command: the command to collect
        """

    async def collect_commands(self, commands: List[AntaCommand]) -> None:
        """
        Collect multiple commands.

        Args:
            commands: the commands to collect
        """
        await asyncio.gather(*(self.collect(command=command) for command in commands))

    @abstractmethod
    async def refresh(self) -> None:
        """
        Update attributes of an AntaDevice instance.

        This coroutine must update the following attributes of AntaDevice:
            - `is_online`: When the device IP is reachable and a port can be open
            - `established`: When a command execution succeeds
            - `hw_model`: The hardware model of the device
        """

    async def copy(self, sources: List[Path], destination: Path, direction: Literal["to", "from"] = "from") -> None:
        """
        Copy files to and from the device, usually through SCP.
        It is not mandatory to implement this for a valid AntaDevice subclass.

        Args:
            sources: List of files to copy to or from the device.
            destination: Local or remote destination when copying the files. Can be a folder.
            direction: Defines if this coroutine copies files to or from the device.
        """
        raise NotImplementedError(f"copy() method has not been implemented in {self.__class__.__name__} definition")


class AsyncEOSDevice(AntaDevice):
    """
    Implementation of AntaDevice for EOS using aio-eapi.

    Attributes:
        name: Device name
        is_online: True if the device IP is reachable and a port can be open
        established: True if remote command execution succeeds
        hw_model: Hardware model of the device
        tags: List of tags for this device
    """

    def __init__(  # pylint: disable=R0913
        self,
        host: str,
        username: str,
        password: str,
        name: Optional[str] = None,
        enable: bool = False,
        enable_password: Optional[str] = None,
        port: Optional[int] = None,
        ssh_port: Optional[int] = 22,
        tags: Optional[List[str]] = None,
        timeout: Optional[float] = None,
        insecure: bool = False,
        proto: Literal["http", "https"] = "https",
    ) -> None:
        """
        Constructor of AsyncEOSDevice

        Args:
            host: Device FQDN or IP
            username: Username to connect to eAPI and SSH
            password: Password to connect to eAPI and SSH
            name: Device name
            enable: Device needs privileged access
            enable_password: Password used to gain privileged access on EOS
            port: eAPI port. Defaults to 80 is proto is 'http' or 443 if proto is 'https'.
            ssh_port: SSH port
            tags: List of tags for this device
            timeout: Timeout value in seconds for outgoing connections. Default to 10 secs.
            insecure: Disable SSH Host Key validation
            proto: eAPI protocol. Value can be 'http' or 'https'
        """
        if name is None:
            name = f"{host}{f':{port}' if port else ''}"
        super().__init__(name, tags)
        self.enable = enable
        self._enable_password = enable_password
        self._session: Device = Device(host=host, port=port, username=username, password=password, proto=proto, timeout=timeout)
        ssh_params: Dict[str, Any] = {}
        if insecure:
            ssh_params.update({"known_hosts": None})
        self._ssh_opts: SSHClientConnectionOptions = SSHClientConnectionOptions(host=host, port=ssh_port, username=username, password=password, **ssh_params)

    def __rich_repr__(self) -> Iterator[Tuple[str, Any]]:
        """
        Implements Rich Repr Protocol
        https://rich.readthedocs.io/en/stable/pretty.html#rich-repr-protocol
        """
        yield from super().__rich_repr__()
        yield "host", self._session.host
        yield "eapi_port", self._session.port
        yield "username", self._ssh_opts.username
        yield "password", self._ssh_opts.password
        yield "enable", self.enable
        yield "enable_password", self._enable_password
        yield "insecure", self._ssh_opts.known_hosts is None
        if __DEBUG__:
            yield "_session", vars(self._session)
            yield "_ssh_opts", vars(self._ssh_opts)

    def __eq__(self, other: object) -> bool:
        """
        Two AsyncEOSDevice objects are equal if the hostname and the port are the same.
        This covers the use case of port forwarding when the host is localhost and the devices have different ports.
        """
        if not isinstance(other, AsyncEOSDevice):
            return False
        return self._session.host == other._session.host and self._session.port == other._session.port

    async def collect(self, command: AntaCommand) -> None:
        """
        Collect device command output from EOS using aio-eapi.

        Supports outformat `json` and `text` as output structure.
        Gain privileged access using the `enable_password` attribute
        of the `AntaDevice` instance if populated.

        Args:
            command: the command to collect
        """
        try:
            commands = []
            if self.enable and self._enable_password is not None:
                commands.append(
                    {
                        "cmd": "enable",
                        "input": str(self._enable_password),
                    }
                )
            elif self.enable:
                # No password
                commands.append({"cmd": "enable"})
            if command.revision:
                commands.append({"cmd": command.command, "revision": command.revision})
            else:
                commands.append({"cmd": command.command})
            response = await self._session.cli(
                commands=commands,
                ofmt=command.ofmt,
                version=command.version,
            )
            # remove first dict related to enable command
            # only applicable to json output
            if command.ofmt in ["json", "text"]:
                # selecting only our command output
                response = response[-1]
            command.output = response
            logger.debug(f"{self.name}: {command}")

        except EapiCommandError as e:
            message = f"Command '{command.command}' failed on {self.name}"
            anta_log_exception(e, message, logger)
            command.failed = e
        except (HTTPError, ConnectError) as e:
            message = f"Cannot connect to device {self.name}"
            anta_log_exception(e, message, logger)
            command.failed = e
        except Exception as e:  # pylint: disable=broad-exception-caught
            message = f"Exception raised while collecting command '{command.command}' on device {self.name}"
            anta_log_exception(e, message, logger)
            command.failed = e
            logger.debug(command)

    async def refresh(self) -> None:
        """
        Update attributes of an AsyncEOSDevice instance.

        This coroutine must update the following attributes of AsyncEOSDevice:
        - is_online: When a device IP is reachable and a port can be open
        - established: When a command execution succeeds
        - hw_model: The hardware model of the device
        """
        # Refresh command
        COMMAND: str = "show version"
        # Hardware model definition in show version
        HW_MODEL_KEY: str = "modelName"
        logger.debug(f"Refreshing device {self.name}")
        self.is_online = await self._session.check_connection()
        if self.is_online:
            try:
                response = await self._session.cli(command=COMMAND)
            except EapiCommandError as e:
                logger.warning(f"Cannot get hardware information from device {self.name}: {e.errmsg}")
            except (HTTPError, ConnectError) as e:
                logger.warning(f"Cannot get hardware information from device {self.name}: {exc_to_str(e)}")
            else:
                if HW_MODEL_KEY in response:
                    self.hw_model = response[HW_MODEL_KEY]
                else:
                    logger.warning(f"Cannot get hardware information from device {self.name}: cannot parse '{COMMAND}'")
        else:
            logger.warning(f"Could not connect to device {self.name}: cannot open eAPI port")
        self.established = bool(self.is_online and self.hw_model)

    async def copy(self, sources: List[Path], destination: Path, direction: Literal["to", "from"] = "from") -> None:
        """
        Copy files to and from the device using asyncssh.scp().

        Args:
            sources: List of files to copy to or from the device.
            destination: Local or remote destination when copying the files. Can be a folder.
            direction: Defines if this coroutine copies files to or from the device.
        """
        async with asyncssh.connect(
            host=self._ssh_opts.host,
            port=self._ssh_opts.port,
            tunnel=self._ssh_opts.tunnel,
            family=self._ssh_opts.family,
            local_addr=self._ssh_opts.local_addr,
            options=self._ssh_opts,
        ) as conn:
            src: Union[List[Tuple[SSHClientConnection, Path]], List[Path]]
            dst: Union[Tuple[SSHClientConnection, Path], Path]
            if direction == "from":
                src = [(conn, file) for file in sources]
                dst = destination
                for file in sources:
                    logger.info(f"Copying '{file}' from device {self.name} to '{destination}' locally")
            elif direction == "to":
                src = sources
                dst = (conn, destination)
                for file in sources:
                    logger.info(f"Copying '{file}' to device {self.name} to '{destination}' remotely")
            else:
                logger.critical(f"'direction' argument to copy() fonction is invalid: {direction}")
                return
            await asyncssh.scp(src, dst)



"""
Test functions related to the EOS software
"""

from typing import List, Optional

from anta.models import AntaCommand, AntaTest


class VerifyEOSVersion(AntaTest):
    """
    Verifies the device is running one of the allowed EOS version.
    """

    name = "VerifyEOSVersion"
    description = "Verifies the device is running one of the allowed EOS version."
    categories = ["software"]
    commands = [AntaCommand(command="show version")]

    @AntaTest.anta_test
    def test(self, versions: Optional[List[str]] = None) -> None:
        """
        Run VerifyEOSVersion validation

        Args:
            versions: List of allowed EOS versions.
        """
        if not versions:
            self.result.is_skipped("VerifyEOSVersion was not run as no versions were given")
            return

        command_output = self.instance_commands[0].json_output

        if command_output["version"] in versions:
            self.result.is_success()
        else:
            self.result.is_failure(f'device is running version {command_output["version"]} not in expected versions: {versions}')


class VerifyTerminAttrVersion(AntaTest):
    """
    Verifies the device is running one of the allowed TerminAttr version.
    """

    name = "VerifyTerminAttrVersion"
    description = "Verifies the device is running one of the allowed TerminAttr version."
    categories = ["software"]
    commands = [AntaCommand(command="show version detail")]

    @AntaTest.anta_test
    def test(self, versions: Optional[List[str]] = None) -> None:
        """
        Run VerifyTerminAttrVersion validation

        Args:
            versions: List of allowed TerminAttr versions.
        """

        if not versions:
            self.result.is_skipped("VerifyTerminAttrVersion was not run as no versions were given")
            return

        command_output = self.instance_commands[0].json_output

        command_output_data = command_output["details"]["packages"]["TerminAttr-core"]["version"]
        if command_output_data in versions:
            self.result.is_success()
        else:
            self.result.is_failure(f"device is running TerminAttr version {command_output_data} and is not in the allowed list: {versions}")


class VerifyEOSExtensions(AntaTest):
    """
    Verifies all EOS extensions installed on the device are enabled for boot persistence.
    """

    name = "VerifyEOSExtensions"
    description = "Verifies all EOS extensions installed on the device are enabled for boot persistence."
    categories = ["software"]
    commands = [AntaCommand(command="show extensions"), AntaCommand(command="show boot-extensions")]

    @AntaTest.anta_test
    def test(self) -> None:
        """Run VerifyEOSExtensions validation"""

        boot_extensions = []

        show_extensions_command_output = self.instance_commands[0].json_output
        show_boot_extensions_command_output = self.instance_commands[1].json_output

        installed_extensions = [
            extension for extension, extension_data in show_extensions_command_output["extensions"].items() if extension_data["status"] == "installed"
        ]

        for extension in show_boot_extensions_command_output["extensions"]:
            extension = extension.strip("\n")
            if extension != "":
                boot_extensions.append(extension)

        installed_extensions.sort()
        boot_extensions.sort()
        if installed_extensions == boot_extensions:
            self.result.is_success()
        else:
            self.result.is_failure(f"Missing EOS extensions: installed {installed_extensions} / configured: {boot_extensions}")



"""
ANTA runner function
"""

import asyncio
import itertools
import logging
from typing import Any, Callable, Dict, List, Optional, Tuple

from anta.inventory import AntaInventory
from anta.models import AntaTest
from anta.result_manager import ResultManager
from anta.result_manager.models import TestResult
from anta.tools.misc import anta_log_exception

logger = logging.getLogger(__name__)

# Keys from the ANTA test catalog file tranferred to kwargs for AntaTest initialization.
TEST_CATALOG_PARAMS = [
    "result_overwrite",
    "template_params",
]


async def main(
    manager: ResultManager,
    inventory: AntaInventory,
    tests: List[Tuple[Callable[..., TestResult], Dict[Any, Any]]],
    tags: Optional[List[str]] = None,
    established_only: bool = True,
) -> None:
    """
    Main coroutine to run ANTA.
    Use this as an entrypoint to the test framwork in your script.

    Args:
        manager (ResultManager): ResultManager object to populate with the test results.
        inventory (AntaInventory): AntaInventory object that includes the device(s).
        tests (List[...]): ANTA test catalog. Output of anta.loader.parse_catalog().
        tags (Optional[List[str]]): List of tags to filter devices from the inventory. Defaults to None.
        established_only (bool): Include only established device(s). Defaults to True.

    Example:
        anta.tests.routing.bgp:
        - VerifyBGPIPv4UnicastCount:
            number: 3
            template_params:
                - vrf: default

        anta.tests.connectivity:
        - VerifyReachability:
            result_overwrite:
              categories:
                - "Overwritten category 1"
              description: "Test with overwritten description"
              custom_field: "Test run by John Doe"
            template_params:
              - src: Loopback0
                dst: 10.1.0.1

    Returns:
        any: ResultManager object gets updated with the test results.
    """
    # Accept 6 arguments here
    # pylint: disable=R0913

    await inventory.connect_inventory()

    # asyncio.gather takes an iterator of the function to run concurrently.
    # we get the cross product of the devices and tests to build that iterator.

    coros = []
    for device, test in itertools.product(inventory.get_inventory(established_only=established_only, tags=tags).values(), tests):
        kwargs = {k: v for k, v in test[1].items() if k in TEST_CATALOG_PARAMS}

        if "result_overwrite" in kwargs:
            result_overwrite = kwargs.pop("result_overwrite")
            result_overwrite = {"result_" + k: v for k, v in result_overwrite.items()}
            kwargs.update(result_overwrite)

        test_params = {k: v for k, v in test[1].items() if k not in TEST_CATALOG_PARAMS}

        try:
            # Instantiate AntaTest object
            test_instance = test[0](device=device, **kwargs)
            coros.append(test_instance.test(eos_data=None, **test_params))
        except Exception as e:  # pylint: disable=broad-exception-caught
            message = "Error when creating ANTA tests"
            anta_log_exception(e, message, logger)

    if AntaTest.progress is not None:
        AntaTest.nrfu_task = AntaTest.progress.add_task("Running NRFU Tests...", total=len(coros))

    logger.info("Running ANTA tests...")
    res = await asyncio.gather(*coros, return_exceptions=True)
    for r in res:
        if isinstance(r, Exception):
            message = "Error in main ANTA Runner"
            anta_log_exception(r, message, logger)
            res.remove(r)
    manager.add_test_results(res)



"""
Test functions related to the device interfaces
"""

import re
from typing import Any, Dict, List, Optional

from anta.decorators import skip_on_platforms
from anta.models import AntaCommand, AntaTemplate, AntaTest


class VerifyInterfaceUtilization(AntaTest):
    """
    Verifies interfaces utilization is below 75%.
    """

    name = "VerifyInterfaceUtilization"
    description = "Verifies interfaces utilization is below 75%."
    categories = ["interfaces"]
    # TODO - move from text to json if possible
    commands = [AntaCommand(command="show interfaces counters rates", ofmt="text")]

    @AntaTest.anta_test
    def test(self) -> None:
        """Run VerifyInterfaceUtilization validation"""

        command_output = self.instance_commands[0].text_output

        wrong_interfaces = {}
        for line in command_output.split("\n")[1:]:
            if len(line) > 0:
                if line.split()[-5] == "-" or line.split()[-2] == "-":
                    pass
                elif float(line.split()[-5].replace("%", "")) > 75.0:
                    wrong_interfaces[line.split()[0]] = line.split()[-5]
                elif float(line.split()[-2].replace("%", "")) > 75.0:
                    wrong_interfaces[line.split()[0]] = line.split()[-2]

        if not wrong_interfaces:
            self.result.is_success()
        else:
            self.result.is_failure(f"The following interfaces have a usage > 75%: {wrong_interfaces}")


class VerifyInterfaceErrors(AntaTest):
    """
    Verifies interfaces error counters are equal to zero.
    """

    name = "VerifyInterfaceErrors"
    description = "Verifies interfaces error counters are equal to zero."
    categories = ["interfaces"]
    commands = [AntaCommand(command="show interfaces counters errors")]

    @AntaTest.anta_test
    def test(self) -> None:
        """Run VerifyInterfaceUtilization validation"""

        command_output = self.instance_commands[0].json_output

        wrong_interfaces: List[Dict[str, Dict[str, int]]] = []
        for interface, outer_v in command_output["interfaceErrorCounters"].items():
            wrong_interfaces.extend({interface: outer_v} for counter, value in outer_v.items() if value > 0)
        if not wrong_interfaces:
            self.result.is_success()
        else:
            self.result.is_failure(f"The following interfaces have non 0 error counter(s): {wrong_interfaces}")


class VerifyInterfaceDiscards(AntaTest):
    """
    Verifies interfaces packet discard counters are equal to zero.
    """

    name = "VerifyInterfaceDiscards"
    description = "Verifies interfaces packet discard counters are equal to zero."
    categories = ["interfaces"]
    commands = [AntaCommand(command="show interfaces counters discards")]

    @AntaTest.anta_test
    def test(self) -> None:
        """Run VerifyInterfaceDiscards validation"""

        command_output = self.instance_commands[0].json_output

        wrong_interfaces: List[Dict[str, Dict[str, int]]] = []

        for interface, outer_v in command_output["interfaces"].items():
            wrong_interfaces.extend({interface: outer_v} for counter, value in outer_v.items() if value > 0)
        if not wrong_interfaces:
            self.result.is_success()
        else:
            self.result.is_failure(f"The following interfaces have non 0 discard counter(s): {wrong_interfaces}")


class VerifyInterfaceErrDisabled(AntaTest):
    """
    Verifies there is no interface in error disable state.
    """

    name = "VerifyInterfaceErrDisabled"
    description = "Verifies there is no interface in error disable state."
    categories = ["interfaces"]
    commands = [AntaCommand(command="show interfaces status")]

    @AntaTest.anta_test
    def test(self) -> None:
        """Run VerifyInterfaceErrDisabled validation"""

        command_output = self.instance_commands[0].json_output

        errdisabled_interfaces = [interface for interface, value in command_output["interfaceStatuses"].items() if value["linkStatus"] == "errdisabled"]

        if errdisabled_interfaces:
            self.result.is_failure(f"The following interfaces are in error disabled state: {errdisabled_interfaces}")
        else:
            self.result.is_success()


class VerifyInterfacesStatus(AntaTest):
    """
    Verifies the number of Ethernet interfaces up/up on the device is higher or equal than a value.
    """

    name = "VerifyInterfacesStatus"
    description = "Verifies the number of Ethernet interfaces up/up on the device is higher or equal than a value."
    categories = ["interfaces"]
    commands = [AntaCommand(command="show interfaces description")]

    @AntaTest.anta_test
    def test(self, minimum: Optional[int] = None) -> None:
        """
        Run VerifyInterfacesStatus validation

        Args:
            minimum: Expected minimum number of Ethernet interfaces up/up.
        """

        if minimum is None or minimum < 0:
            self.result.is_skipped(f"VerifyInterfacesStatus was not run as an invalid minimum value was given {minimum}.")
            return

        command_output = self.instance_commands[0].json_output

        count_up_up = 0
        other_ethernet_interfaces = []

        for interface in command_output["interfaceDescriptions"]:
            interface_dict = command_output["interfaceDescriptions"][interface]
            if "Ethernet" in interface:
                if re.match(r"connected|up", interface_dict["lineProtocolStatus"]) and re.match(r"connected|up", interface_dict["interfaceStatus"]):
                    count_up_up += 1
                else:
                    other_ethernet_interfaces.append(interface)

        if count_up_up >= minimum:
            self.result.is_success()
        else:
            self.result.is_failure(f"Only {count_up_up}, less than {minimum} Ethernet interfaces are UP/UP")
            self.result.messages.append(f"The following Ethernet interfaces are not UP/UP: {other_ethernet_interfaces}")


class VerifyStormControlDrops(AntaTest):
    """
    Verifies the device did not drop packets due its to storm-control configuration.
    """

    name = "VerifyStormControlDrops"
    description = "Verifies the device did not drop packets due its to storm-control configuration."
    categories = ["interfaces"]
    commands = [AntaCommand(command="show storm-control")]

    @skip_on_platforms(["cEOSLab", "vEOS-lab"])
    @AntaTest.anta_test
    def test(self) -> None:
        """Run VerifyStormControlDrops validation"""

        command_output = self.instance_commands[0].json_output

        storm_controlled_interfaces: Dict[str, Dict[str, Any]] = {}
        for interface, interface_dict in command_output["interfaces"].items():
            for traffic_type, traffic_type_dict in interface_dict["trafficTypes"].items():
                if "drop" in traffic_type_dict and traffic_type_dict["drop"] != 0:
                    storm_controlled_interface_dict = storm_controlled_interfaces.setdefault(interface, {})
                    storm_controlled_interface_dict.update({traffic_type: traffic_type_dict["drop"]})

        if not storm_controlled_interfaces:
            self.result.is_success()
        else:
            self.result.is_failure(f"The following interfaces have none 0 storm-control drop counters {storm_controlled_interfaces}")


class VerifyPortChannels(AntaTest):
    """
    Verifies there is no inactive port in port channels.
    """

    name = "VerifyPortChannels"
    description = "Verifies there is no inactive port in port channels."
    categories = ["interfaces"]
    commands = [AntaCommand(command="show port-channel")]

    @skip_on_platforms(["cEOSLab", "vEOS-lab"])
    @AntaTest.anta_test
    def test(self) -> None:
        """Run VerifyPortChannels validation"""

        command_output = self.instance_commands[0].json_output

        po_with_invactive_ports: List[Dict[str, str]] = []
        for portchannel, portchannel_dict in command_output["portChannels"].items():
            if len(portchannel_dict["inactivePorts"]) != 0:
                po_with_invactive_ports.extend({portchannel: portchannel_dict["inactivePorts"]})

        if not po_with_invactive_ports:
            self.result.is_success()
        else:
            self.result.is_failure(f"The following port-channels have inactive port(s): {po_with_invactive_ports}")


class VerifyIllegalLACP(AntaTest):
    """
    Verifies there is no illegal LACP packets received.
    """

    name = "VerifyIllegalLACP"
    description = "Verifies there is no illegal LACP packets received."
    categories = ["interfaces"]
    commands = [AntaCommand(command="show lacp counters all-ports")]

    @AntaTest.anta_test
    def test(self) -> None:
        """Run VerifyIllegalLACP validation"""

        command_output = self.instance_commands[0].json_output

        po_with_illegal_lacp: List[Dict[str, Dict[str, int]]] = []
        for portchannel, portchannel_dict in command_output["portChannels"].items():
            po_with_illegal_lacp.extend(
                {portchannel: interface} for interface, interface_dict in portchannel_dict["interfaces"].items() if interface_dict["illegalRxCount"] != 0
            )

        if not po_with_illegal_lacp:
            self.result.is_success()
        else:
            self.result.is_failure("The following port-channels have recieved illegal lacp packets on the " f"following ports: {po_with_illegal_lacp}")


class VerifyLoopbackCount(AntaTest):
    """
    Verifies the number of loopback interfaces on the device is the one we expect and if none of the loopback is down.
    """

    name = "VerifyLoopbackCount"
    description = "Verifies the number of loopback interfaces on the device is the one we expect and if none of the loopback is down."
    categories = ["interfaces"]
    commands = [AntaCommand(command="show ip interface brief")]

    @AntaTest.anta_test
    def test(self, number: Optional[int] = None) -> None:
        """
        Run VerifyLoopbackCount validation

        Args:
            number: Number of loopback interfaces expected to be present.
        """

        if number is None:
            self.result.is_skipped("VerifyLoopbackCount was not run as no number value was given.")
            return

        command_output = self.instance_commands[0].json_output

        loopback_count = 0
        down_loopback_interfaces = []

        for interface in command_output["interfaces"]:
            interface_dict = command_output["interfaces"][interface]
            if "Loopback" in interface:
                loopback_count += 1
                if not (interface_dict["lineProtocolStatus"] == "up" and interface_dict["interfaceStatus"] == "connected"):
                    down_loopback_interfaces.append(interface)

        if loopback_count == number and len(down_loopback_interfaces) == 0:
            self.result.is_success()
        else:
            self.result.is_failure()
            if loopback_count != number:
                self.result.is_failure(f"Found {loopback_count} Loopbacks when expecting {number}")
            elif len(down_loopback_interfaces) != 0:
                self.result.is_failure(f"The following Loopbacks are not up: {down_loopback_interfaces}")


class VerifySVI(AntaTest):
    """
    Verifies there is no interface vlan down.
    """

    name = "VerifySVI"
    description = "Verifies there is no interface vlan down."
    categories = ["interfaces"]
    commands = [AntaCommand(command="show ip interface brief")]

    @AntaTest.anta_test
    def test(self) -> None:
        """Run VerifySVI validation"""

        command_output = self.instance_commands[0].json_output

        down_svis = []

        for interface in command_output["interfaces"]:
            interface_dict = command_output["interfaces"][interface]
            if "Vlan" in interface:
                if not (interface_dict["lineProtocolStatus"] == "up" and interface_dict["interfaceStatus"] == "connected"):
                    down_svis.append(interface)

        if len(down_svis) == 0:
            self.result.is_success()
        else:
            self.result.is_failure(f"The following SVIs are not up: {down_svis}")


class VerifyL3MTU(AntaTest):
    """
    Verifies the global layer 3 Maximum Transfer Unit (MTU) for all layer 3 interfaces.

    Expected Results:
        * success: The test will pass if all layer 3 interfaces have the proper MTU configured.
        * failure: The test will fail if one or many layer 3 interfaces have the wrong MTU configured.
        * skipped: The test will be skipped if the MTU value is not provided.

    Limitations:
        * Only Ethernet, Port-Channel, Vlan interfaces are supported.
        * Other interface types, like Management, Loopback, Vxlan, Tunnel are currently not supported.

    https://www.arista.com/en/support/toi/eos-4-23-1f/14388-global-knob-to-set-mtu-for-all-layer-3-interfaces

    """

    name = "VerifyL3MTU"
    description = "Verifies the global layer 3 Maximum Transfer Unit (MTU) for all layer 3 interfaces."
    categories = ["interfaces"]
    commands = [AntaCommand(command="show interfaces")]

    NOT_SUPPORTED_INTERFACES: List[str] = ["Management", "Loopback", "Vxlan", "Tunnel"]

    @AntaTest.anta_test
    def test(self, mtu: int = 1500) -> None:
        """
        Run VerifyL3MTU validation

        Args:
          mtu: Layer 3 MTU to verify. Defaults to 1500.

        """

        if not mtu:
            self.result.is_skipped(f"{self.__class__.name} did not run because mtu was not supplied")
            return

        command_output = self.instance_commands[0].json_output

        wrong_l3mtu_intf = []

        for interface, values in command_output["interfaces"].items():
            if re.sub(r"\d+$", "", interface) not in self.NOT_SUPPORTED_INTERFACES:
                if values["forwardingModel"] == "routed" and values["mtu"] != mtu:
                    wrong_l3mtu_intf.append(interface)

        if not wrong_l3mtu_intf:
            self.result.is_success()

        else:
            self.result.is_failure(f"The following interface(s) have the wrong MTU configured: {wrong_l3mtu_intf}")


class VerifyIPProxyARP(AntaTest):
    """
    Verifies if Proxy-ARP is enabled for the provided list of interface(s).

    Expected Results:
        * success: The test will pass if Proxy-ARP is enabled on the specified interface(s).
        * failure: The test will fail if Proxy-ARP is disabled on the specified interface(s).
        * error: The test will give an error if a list of interface(s) is not provided as template_params.

    """

    name = "VerifyIPProxyARP"
    description = "Verifies if Proxy-ARP is enabled for the provided list of interface(s)."
    categories = ["interfaces"]
    template = AntaTemplate(template="show ip interface {intf}")

    @AntaTest.anta_test
    def test(self) -> None:
        """
        Run VerifyIPProxyARP validation.
        """

        disabled_intf = []
        for command in self.instance_commands:
            if command.params and "intf" in command.params:
                intf = command.params["intf"]
            if not command.json_output["interfaces"][intf]["proxyArp"]:
                disabled_intf.append(intf)

        if disabled_intf:
            self.result.is_failure(f"The following interface(s) have Proxy-ARP disabled: {disabled_intf}")

        else:
            self.result.is_success()



"""
Test functions related to the EOS various SNMP settings
"""
from __future__ import annotations

from typing import Optional

from anta.models import AntaCommand, AntaTest


class VerifySnmpStatus(AntaTest):
    """
    Verifies whether the SNMP agent is enabled in a specified VRF.

    Expected Results:
        * success: The test will pass if the SNMP agent is enabled in the specified VRF.
        * failure: The test will fail if the SNMP agent is disabled in the specified VRF.
        * skipped: The test will be skipped if the VRF parameter is not provided.
    """

    name = "VerifySnmpStatus"
    description = "Verifies if the SNMP agent is enabled."
    categories = ["snmp"]
    commands = [AntaCommand(command="show snmp")]

    @AntaTest.anta_test
    def test(self, vrf: str = "default") -> None:
        """
        Run VerifySnmpStatus validation.

        Args:
            vrf: The name of the VRF in which to check for the SNMP agent. Defaults to 'default'.
        """
        if not vrf:
            self.result.is_skipped(f"{self.__class__.name} did not run because vrf was not supplied")
        else:
            command_output = self.instance_commands[0].json_output

            if command_output["enabled"] and vrf in command_output["vrfs"]["snmpVrfs"]:
                self.result.is_success()
            else:
                self.result.is_failure(f"SNMP agent disabled in vrf {vrf}")


class VerifySnmpIPv4Acl(AntaTest):
    """
    Verifies if the SNMP agent has the right number IPv4 ACL(s) configured for a specified VRF.

    Expected results:
        * success: The test will pass if the SNMP agent has the provided number of IPv4 ACL(s) in the specified VRF.
        * failure: The test will fail if the SNMP agent has not the right number of IPv4 ACL(s) in the specified VRF.
        * skipped: The test will be skipped if the number of IPv4 ACL(s) or VRF parameter is not provided.
    """

    name = "VerifySnmpIPv4Acl"
    description = "Verifies if the SNMP agent has IPv4 ACL(s) configured."
    categories = ["snmp"]
    commands = [AntaCommand(command="show snmp ipv4 access-list summary")]

    @AntaTest.anta_test
    def test(self, number: Optional[int] = None, vrf: str = "default") -> None:
        """
        Run VerifySnmpIPv4Acl validation.

        Args:
            number: The number of expected IPv4 ACL(s).
            vrf: The name of the VRF in which to check for the SNMP agent. Defaults to 'default'.
        """
        if not number or not vrf:
            self.result.is_skipped(f"{self.__class__.name} did not run because number or vrf was not supplied")
            return

        command_output = self.instance_commands[0].json_output

        ipv4_acl_list = command_output["ipAclList"]["aclList"]
        ipv4_acl_number = len(ipv4_acl_list)
        not_configured_acl_list = []

        if ipv4_acl_number != number:
            self.result.is_failure(f"Expected {number} SNMP IPv4 ACL(s) in vrf {vrf} but got {ipv4_acl_number}")
            return

        for ipv4_acl in ipv4_acl_list:
            if vrf not in ipv4_acl["configuredVrfs"] or vrf not in ipv4_acl["activeVrfs"]:
                not_configured_acl_list.append(ipv4_acl["name"])

        if not_configured_acl_list:
            self.result.is_failure(f"SNMP IPv4 ACL(s) not configured or active in vrf {vrf}: {not_configured_acl_list}")
        else:
            self.result.is_success()


class VerifySnmpIPv6Acl(AntaTest):
    """
    Verifies if the SNMP agent has the right number IPv6 ACL(s) configured for a specified VRF.

    Expected results:
        * success: The test will pass if the SNMP agent has the provided number of IPv6 ACL(s) in the specified VRF.
        * failure: The test will fail if the SNMP agent has not the right number of IPv6 ACL(s) in the specified VRF.
        * skipped: The test will be skipped if the number of IPv6 ACL(s) or VRF parameter is not provided.
    """

    name = "VerifySnmpIPv6Acl"
    description = "Verifies if the SNMP agent has IPv6 ACL(s) configured."
    categories = ["snmp"]
    commands = [AntaCommand(command="show snmp ipv6 access-list summary")]

    @AntaTest.anta_test
    def test(self, number: Optional[int] = None, vrf: str = "default") -> None:
        """
        Run VerifySnmpIPv6Acl validation.

        Args:
            number: The number of expected IPv6 ACL(s).
            vrf: The name of the VRF in which to check for the SNMP agent. Defaults to 'default'.
        """
        if not number or not vrf:
            self.result.is_skipped(f"{self.__class__.name} did not run because number or vrf was not supplied")
            return

        command_output = self.instance_commands[0].json_output

        ipv6_acl_list = command_output["ipv6AclList"]["aclList"]
        ipv6_acl_number = len(ipv6_acl_list)
        not_configured_acl_list = []

        if ipv6_acl_number != number:
            self.result.is_failure(f"Expected {number} SNMP IPv6 ACL(s) in vrf {vrf} but got {ipv6_acl_number}")
            return

        for ipv6_acl in ipv6_acl_list:
            if vrf not in ipv6_acl["configuredVrfs"] or vrf not in ipv6_acl["activeVrfs"]:
                not_configured_acl_list.append(ipv6_acl["name"])

        if not_configured_acl_list:
            self.result.is_failure(f"SNMP IPv6 ACL(s) not configured or active in vrf {vrf}: {not_configured_acl_list}")
        else:
            self.result.is_success()



"""
Test functions related to the EOS various AAA settings
"""
from __future__ import annotations

from typing import List, Optional

from anta.models import AntaCommand, AntaTest


def _check_group_methods(methods: List[str]) -> List[str]:
    """
    Verifies if the provided methods in various AAA tests start with 'group'.

    Args:
        methods: List of AAA methods. Methods should be in the right order.
    """
    built_in_methods = ["local", "none", "logging"]

    return [f"group {method}" if method not in built_in_methods and not method.startswith("group ") else method for method in methods]


def _check_auth_type(auth_types: List[str], valid_auth_types: List[str]) -> None:
    """
    Verifies if the provided auth types in various AAA tests are valid.

    Args:
        auth_types: List of AAA auth types to validate.
        valid_auth_types: List of valid AAA auth types to validate against.
    """
    if len(auth_types) > len(valid_auth_types):
        raise ValueError(f"Too many parameters provided in auth_types. Valid parameters are: {valid_auth_types}")

    for auth_type in auth_types:
        if auth_type not in valid_auth_types:
            raise ValueError(f"Wrong parameter provided in auth_types. Valid parameters are: {valid_auth_types}")


class VerifyTacacsSourceIntf(AntaTest):
    """
    Verifies TACACS source-interface for a specified VRF.

    Expected Results:
        * success: The test will pass if the provided TACACS source-interface is configured in the specified VRF.
        * failure: The test will fail if the provided TACACS source-interface is NOT configured in the specified VRF.
        * skipped: The test will be skipped if source-interface or VRF is not provided.
    """

    name = "VerifyTacacsSourceIntf"
    description = "Verifies TACACS source-interface for a specified VRF."
    categories = ["aaa"]
    commands = [AntaCommand(command="show tacacs")]

    @AntaTest.anta_test
    def test(self, intf: Optional[str] = None, vrf: str = "default") -> None:
        """
        Run VerifyTacacsSourceIntf validation.

        Args:
            intf: Source-interface to use as source IP of TACACS messages.
            vrf: The name of the VRF to transport TACACS messages. Defaults to 'default'.
        """
        if not intf or not vrf:
            self.result.is_skipped(f"{self.__class__.name} did not run because intf or vrf was not supplied")
            return

        command_output = self.instance_commands[0].json_output

        try:
            if command_output["srcIntf"][vrf] == intf:
                self.result.is_success()
            else:
                self.result.is_failure(f"Wrong source-interface configured in VRF {vrf}")

        except KeyError:
            self.result.is_failure(f"Source-interface {intf} is not configured in VRF {vrf}")


class VerifyTacacsServers(AntaTest):
    """
    Verifies TACACS servers are configured for a specified VRF.

    Expected Results:
        * success: The test will pass if the provided TACACS servers are configured in the specified VRF.
        * failure: The test will fail if the provided TACACS servers are NOT configured in the specified VRF.
        * skipped: The test will be skipped if TACACS servers or VRF are not provided.
    """

    name = "VerifyTacacsServers"
    description = "Verifies TACACS servers are configured for a specified VRF."
    categories = ["aaa"]
    commands = [AntaCommand(command="show tacacs")]

    @AntaTest.anta_test
    def test(self, servers: Optional[List[str]] = None, vrf: str = "default") -> None:
        """
        Run VerifyTacacsServers validation.

        Args:
            servers: List of TACACS servers IP addresses.
            vrf: The name of the VRF to transport TACACS messages. Defaults to 'default'.
        """
        if not servers or not vrf:
            self.result.is_skipped(f"{self.__class__.name} did not run because servers or vrf were not supplied")
            return

        command_output = self.instance_commands[0].json_output

        tacacs_servers = command_output["tacacsServers"]

        if not tacacs_servers:
            self.result.is_failure("No TACACS servers are configured")
            return

        not_configured = [
            server
            for server in servers
            if not any(server == tacacs_server["serverInfo"]["hostname"] and vrf == tacacs_server["serverInfo"]["vrf"] for tacacs_server in tacacs_servers)
        ]

        if not not_configured:
            self.result.is_success()
        else:
            self.result.is_failure(f"TACACS servers {not_configured} are not configured in VRF {vrf}")


class VerifyTacacsServerGroups(AntaTest):
    """
    Verifies if the provided TACACS server group(s) are configured.

    Expected Results:
        * success: The test will pass if the provided TACACS server group(s) are configured.
        * failure: The test will fail if one or all the provided TACACS server group(s) are NOT configured.
        * skipped: The test will be skipped if TACACS server group(s) are not provided.
    """

    name = "VerifyTacacsServerGroups"
    description = "Verifies if the provided TACACS server group(s) are configured."
    categories = ["aaa"]
    commands = [AntaCommand(command="show tacacs")]

    @AntaTest.anta_test
    def test(self, groups: Optional[List[str]] = None) -> None:
        """
        Run VerifyTacacsServerGroups validation.

        Args:
            groups: List of TACACS server group.
        """
        if not groups:
            self.result.is_skipped(f"{self.__class__.name} did not run because groups were not supplied")
            return

        command_output = self.instance_commands[0].json_output

        tacacs_groups = command_output["groups"]

        if not tacacs_groups:
            self.result.is_failure("No TACACS server group(s) are configured")
            return

        not_configured = [group for group in groups if group not in tacacs_groups]

        if not not_configured:
            self.result.is_success()
        else:
            self.result.is_failure(f"TACACS server group(s) {not_configured} are not configured")


class VerifyAuthenMethods(AntaTest):
    """
    Verifies the AAA authentication method lists for different authentication types (login, enable, dot1x).

    Expected Results:
        * success: The test will pass if the provided AAA authentication method list is matching in the configured authentication types.
        * failure: The test will fail if the provided AAA authentication method list is NOT matching in the configured authentication types.
        * skipped: The test will be skipped if the AAA authentication method list or authentication type list are not provided.
    """

    name = "VerifyAuthenMethods"
    description = "Verifies the AAA authentication method lists for different authentication types (login, enable, dot1x)."
    categories = ["aaa"]
    commands = [AntaCommand(command="show aaa methods authentication")]

    @AntaTest.anta_test
    def test(self, methods: Optional[List[str]] = None, auth_types: Optional[List[str]] = None) -> None:
        """
        Run VerifyAuthenMethods validation.

        Args:
            methods: List of AAA authentication methods. Methods should be in the right order.
            auth_types: List of authentication types to verify. List elements must be: login, enable, dot1x.
        """
        if not methods or not auth_types:
            self.result.is_skipped(f"{self.__class__.name} did not run because methods or auth_types were not supplied")
            return

        methods_with_group = _check_group_methods(methods)

        _check_auth_type(auth_types, ["login", "enable", "dot1x"])

        command_output = self.instance_commands[0].json_output

        not_matching = []

        for auth_type in auth_types:
            auth_type_key = f"{auth_type}AuthenMethods"

            if auth_type_key == "loginAuthenMethods":
                if not command_output[auth_type_key].get("login"):
                    self.result.is_failure("AAA authentication methods are not configured for login console")
                    return

                if command_output[auth_type_key]["login"]["methods"] != methods_with_group:
                    self.result.is_failure(f"AAA authentication methods {methods} are not matching for login console")
                    return

            if command_output[auth_type_key]["default"]["methods"] != methods_with_group:
                not_matching.append(auth_type)

        if not not_matching:
            self.result.is_success()
        else:
            self.result.is_failure(f"AAA authentication methods {methods} are not matching for {not_matching}")


class VerifyAuthzMethods(AntaTest):
    """
    Verifies the AAA authorization method lists for different authorization types (commands, exec).

    Expected Results:
        * success: The test will pass if the provided AAA authorization method list is matching in the configured authorization types.
        * failure: The test will fail if the provided AAA authorization method list is NOT matching in the configured authorization types.
        * skipped: The test will be skipped if the AAA authentication method list or authorization type list are not provided.
    """

    name = "VerifyAuthzMethods"
    description = "Verifies the AAA authorization method lists for different authorization types (commands, exec)."
    categories = ["aaa"]
    commands = [AntaCommand(command="show aaa methods authorization")]

    @AntaTest.anta_test
    def test(self, methods: Optional[List[str]] = None, auth_types: Optional[List[str]] = None) -> None:
        """
        Run VerifyAuthzMethods validation.

        Args:
            methods: List of AAA authorization methods. Methods should be in the right order.
            auth_types: List of authorization types to verify. List elements must be: commands, exec.
        """
        if not methods or not auth_types:
            self.result.is_skipped(f"{self.__class__.name} did not run because methods or auth_types were not supplied")
            return

        _check_auth_type(auth_types, ["commands", "exec"])

        methods_with_group = _check_group_methods(methods)

        command_output = self.instance_commands[0].json_output

        not_matching = []

        for auth_type in auth_types:
            auth_type_key = f"{auth_type}AuthzMethods"

            method_key = list(command_output[auth_type_key].keys())[0]

            if command_output[auth_type_key][method_key]["methods"] != methods_with_group:
                not_matching.append(auth_type)

        if not not_matching:
            self.result.is_success()
        else:
            self.result.is_failure(f"AAA authorization methods {methods} are not matching for {not_matching}")


class VerifyAcctDefaultMethods(AntaTest):
    """
    Verifies the AAA accounting default method lists for different accounting types (system, exec, commands, dot1x).

    Expected Results:
        * success: The test will pass if the provided AAA accounting default method list is matching in the configured accounting types.
        * failure: The test will fail if the provided AAA accounting default method list is NOT matching in the configured accounting types.
        * skipped: The test will be skipped if the AAA accounting default method list or accounting type list are not provided.
    """

    name = "VerifyAcctDefaultMethods"
    description = "Verifies the AAA accounting default method lists for different accounting types (system, exec, commands, dot1x)."
    categories = ["aaa"]
    commands = [AntaCommand(command="show aaa methods accounting")]

    @AntaTest.anta_test
    def test(self, methods: Optional[List[str]] = None, auth_types: Optional[List[str]] = None) -> None:
        """
        Run VerifyAcctDefaultMethods validation.

        Args:
            methods: List of AAA accounting default methods. Methods should be in the right order.
            auth_types: List of accounting types to verify. List elements must be: commands, exec, system, dot1x.
        """
        if not methods or not auth_types:
            self.result.is_skipped(f"{self.__class__.name} did not run because methods or auth_types were not supplied")
            return

        methods_with_group = _check_group_methods(methods)

        _check_auth_type(auth_types, ["system", "exec", "commands", "dot1x"])

        command_output = self.instance_commands[0].json_output

        not_matching = []
        not_configured = []

        for auth_type in auth_types:
            auth_type_key = f"{auth_type}AcctMethods"

            method_key = list(command_output[auth_type_key].keys())[0]

            if not command_output[auth_type_key][method_key].get("defaultAction"):
                not_configured.append(auth_type)

            if command_output[auth_type_key][method_key]["defaultMethods"] != methods_with_group:
                not_matching.append(auth_type)

        if not_configured:
            self.result.is_failure(f"AAA default accounting is not configured for {not_configured}")
            return

        if not not_matching:
            self.result.is_success()
        else:
            self.result.is_failure(f"AAA accounting default methods {methods} are not matching for {not_matching}")


class VerifyAcctConsoleMethods(AntaTest):
    """
    Verifies the AAA accounting console method lists for different accounting types (system, exec, commands, dot1x).

    Expected Results:
        * success: The test will pass if the provided AAA accounting console method list is matching in the configured accounting types.
        * failure: The test will fail if the provided AAA accounting console method list is NOT matching in the configured accounting types.
        * skipped: The test will be skipped if the AAA accounting console method list or accounting type list are not provided.
    """

    name = "VerifyAcctConsoleMethods"
    description = "Verifies the AAA accounting console method lists for different accounting types (system, exec, commands, dot1x)."
    categories = ["aaa"]
    commands = [AntaCommand(command="show aaa methods accounting")]

    @AntaTest.anta_test
    def test(self, methods: Optional[List[str]] = None, auth_types: Optional[List[str]] = None) -> None:
        """
        Run VerifyAcctConsoleMethods validation.

        Args:
            methods: List of AAA accounting console methods. Methods should be in the right order.
            auth_types: List of accounting types to verify. List elements must be: commands, exec, system, dot1x.
        """
        if not methods or not auth_types:
            self.result.is_skipped(f"{self.__class__.name} did not run because methods or auth_types were not supplied")
            return

        methods_with_group = _check_group_methods(methods)

        _check_auth_type(auth_types, ["system", "exec", "commands", "dot1x"])

        command_output = self.instance_commands[0].json_output

        not_matching = []
        not_configured = []

        for auth_type in auth_types:
            auth_type_key = f"{auth_type}AcctMethods"

            method_key = list(command_output[auth_type_key].keys())[0]

            if not command_output[auth_type_key][method_key].get("consoleAction"):
                not_configured.append(auth_type)

            if command_output[auth_type_key][method_key]["consoleMethods"] != methods_with_group:
                not_matching.append(auth_type)

        if not_configured:
            self.result.is_failure(f"AAA console accounting is not configured for {not_configured}")
            return

        if not not_matching:
            self.result.is_success()
        else:
            self.result.is_failure(f"AAA accounting console methods {methods} are not matching for {not_matching}")



#!/usr/bin/env python
# coding: utf-8 -*-
# pylint: disable = redefined-outer-name

"""
Commands for Anta CLI to run check commands.
"""

import asyncio
import json
import logging
import os
import sys
from pathlib import Path
from typing import List, Optional

import click
from cvprac.cvp_client import CvpClient
from cvprac.cvp_client_errors import CvpApiError
from rich.pretty import pretty_repr

from anta.cli.console import console
from anta.cli.utils import parse_tags
from anta.models import DEFAULT_TAG

from .utils import create_inventory_from_ansible, create_inventory_from_cvp, get_cv_token

logger = logging.getLogger(__name__)


@click.command(no_args_is_help=True)
@click.option("--cvp-ip", "-ip", default=None, help="CVP IP Address", type=str, required=True)
@click.option("--cvp-username", "-u", default=None, help="CVP Username", type=str, required=True)
@click.option("--cvp-password", "-p", default=None, help="CVP Password / token", type=str, required=True)
@click.option("--cvp-container", "-c", default=None, help="Container where devices are configured", type=str, required=False)
@click.option("--inventory-directory", "-d", default="anta_inventory", help="Path to save inventory file", type=click.Path())
def from_cvp(inventory_directory: str, cvp_ip: str, cvp_username: str, cvp_password: str, cvp_container: str) -> None:
    """
    Build ANTA inventory from Cloudvision

    TODO - handle get_inventory and get_devices_in_container failure
    """
    # pylint: disable=too-many-arguments
    logger.info(f"Getting auth token from {cvp_ip} for user {cvp_username}")
    token = get_cv_token(cvp_ip=cvp_ip, cvp_username=cvp_username, cvp_password=cvp_password)

    # Create output directory
    cwd = os.getcwd()
    out_dir = os.path.dirname(f"{cwd}/{inventory_directory}/")
    if not os.path.exists(out_dir):
        logger.info(f"Creating inventory folder {out_dir}")
        os.makedirs(out_dir)

    clnt = CvpClient()
    print(clnt)
    try:
        clnt.connect(nodes=[cvp_ip], username="", password="", api_token=token)
    except CvpApiError as error:
        logger.error(f"Error connecting to cvp: {error}")
        sys.exit(1)
    logger.info(f"Connected to CVP {cvp_ip}")

    cvp_inventory = None
    if cvp_container is None:
        # Get a list of all devices
        logger.info(f"Getting full inventory from {cvp_ip}")
        cvp_inventory = clnt.api.get_inventory()
    else:
        # Get devices under a container
        logger.info(f"Getting inventory for container {cvp_container} from {cvp_ip}")
        cvp_inventory = clnt.api.get_devices_in_container(cvp_container)
    create_inventory_from_cvp(cvp_inventory, out_dir, cvp_container)


@click.command(no_args_is_help=True)
@click.option("--ansible-group", "-g", help="Ansible group to filter", type=str, required=False)
@click.option(
    "--ansible-inventory",
    "-i",
    default=None,
    help="Path to your ansible inventory file to read",
    type=click.Path(file_okay=True, dir_okay=False, exists=True, path_type=Path),
)
@click.option(
    "--output",
    "-o",
    default="inventory-ansible.yml",
    help="Path to save inventory file",
    type=click.Path(file_okay=True, dir_okay=False, exists=False, writable=True, path_type=Path),
)
def from_ansible(output: Path, ansible_inventory: Path, ansible_group: str) -> None:
    """Build ANTA inventory from an ansible inventory YAML file"""
    logger.info(f"Building inventory from ansible file {ansible_inventory}")

    # Create output directory
    output.parent.mkdir(parents=True, exist_ok=True)
    create_inventory_from_ansible(
        inventory=ansible_inventory,
        output_file=output,
        ansible_root=ansible_group,
    )


@click.command()
@click.pass_context
@click.option("--tags", "-t", help="List of tags using comma as separator: tag1,tag2,tag3", type=str, required=False, callback=parse_tags)
@click.option("--connected/--not-connected", help="Display inventory after connection has been created", default=False, required=False)
def inventory(ctx: click.Context, tags: Optional[List[str]], connected: bool) -> None:
    """Show inventory loaded in ANTA."""

    logger.debug(f"Requesting devices for tags: {tags}")
    console.print("Current inventory content is:", style="white on blue")

    if connected:
        asyncio.run(ctx.obj["inventory"].connect_inventory())

    inventory_result = ctx.obj["inventory"].get_inventory(tags=tags)
    console.print(pretty_repr(inventory_result))


@click.command()
@click.pass_context
def tags(ctx: click.Context) -> None:
    """Get list of configured tags in user inventory."""
    tags_found = []
    for device in ctx.obj["inventory"].values():
        tags_found += device.tags
    tags_found = sorted(set(tags_found))
    console.print("Tags found:")
    console.print_json(json.dumps(tags_found, indent=2))
    console.print(f"\n* note that tag [green]{DEFAULT_TAG}[/green] has been added by anta")



"""
Toolkit for ANTA.
"""

from __future__ import annotations

import logging
import traceback
from typing import Optional

from anta import __DEBUG__

logger = logging.getLogger(__name__)


def anta_log_exception(exception: Exception, message: Optional[str] = None, calling_logger: Optional[logging.Logger] = None) -> None:
    """
    Helper function to help log exceptions:
    * if anta.__DEBUG__ is True then the logger.exception method is called to get the traceback
    * otherwise logger.error is called

    Args:
        exception (Exception): The Exception being logged
        message (str): An optional message
        calling_logger (logging.Logger): A logger to which the exception should be logged
                                         if not present, the logger in this file is used.

    """
    if calling_logger is None:
        calling_logger = logger
    if __DEBUG__:
        calling_logger.exception(message)
    else:
        log_message = exc_to_str(exception)
        if message is not None:
            log_message = f"{message} {log_message}"
        calling_logger.error(log_message)


def exc_to_str(exception: Exception) -> str:
    """
    Helper function that returns a human readable string from an Exception object
    """
    res = f"{type(exception).__name__}"
    if str(exception):
        res += f" ({str(exception)})"
    elif hasattr(exception, "errmsg"):
        # TODO - remove when we bump aio-eapi once our PR is merged there
        res += f" ({exception.errmsg})"
    return res


def tb_to_str(exception: Exception) -> str:
    """
    Helper function that returns a traceback string from an Exception object
    """
    return "Traceback (most recent call last):\n" + "".join(traceback.format_tb(exception.__traceback__))



#!/usr/bin/env python
# coding: utf-8 -*-

"""
Commands for Anta CLI to execute EOS commands.
"""

import asyncio
import logging
import sys
from datetime import datetime
from pathlib import Path
from typing import List, Optional

import click
from yaml import safe_load

from anta.cli.exec.utils import clear_counters_utils, collect_commands, collect_scheduled_show_tech
from anta.cli.utils import parse_tags

logger = logging.getLogger(__name__)


@click.command()
@click.pass_context
@click.option("--tags", "-t", help="List of tags using comma as separator: tag1,tag2,tag3", type=str, required=False, callback=parse_tags)
def clear_counters(ctx: click.Context, tags: Optional[List[str]]) -> None:
    """Clear counter statistics on EOS devices"""
    asyncio.run(clear_counters_utils(ctx.obj["inventory"], tags=tags))


@click.command()
@click.pass_context
@click.option("--tags", "-t", help="List of tags using comma as separator: tag1,tag2,tag3", type=str, required=False, callback=parse_tags)
@click.option(
    "--commands-list",
    "-c",
    help="File with list of commands to collect",
    required=True,
    show_envvar=True,
    type=click.Path(file_okay=True, dir_okay=False, exists=True, readable=True, path_type=Path),
)
@click.option(
    "--output",
    "-o",
    show_envvar=True,
    type=click.Path(file_okay=False, dir_okay=True, exists=False, writable=True, path_type=Path),
    help="Directory to save commands output.",
    default=f"anta_snapshot_{datetime.now().strftime('%Y-%m-%d_%H_%M_%S')}",
    show_default=True,
)
def snapshot(ctx: click.Context, tags: Optional[List[str]], commands_list: Path, output: Path) -> None:
    """Collect commands output from devices in inventory"""
    print(f"Collecting data for {commands_list}")
    print(f"Output directory is {output}")
    try:
        with open(commands_list, "r", encoding="UTF-8") as file:
            file_content = file.read()
            eos_commands = safe_load(file_content)
    except FileNotFoundError:
        logger.error(f"Error reading {commands_list}")
        sys.exit(1)
    asyncio.run(collect_commands(ctx.obj["inventory"], eos_commands, output, tags=tags))


@click.command()
@click.pass_context
@click.option("--output", "-o", default="./tech-support", show_default=True, help="Path for tests catalog", type=click.Path(path_type=Path), required=False)
@click.option("--latest", help="Number of scheduled show-tech to retrieve", type=int, required=False)
@click.option(
    "--configure",
    help="Ensure devices have 'aaa authorization exec default local' configured (required for SCP on EOS). THIS WILL CHANGE THE CONFIGURATION OF YOUR NETWORK.",
    default=False,
    is_flag=True,
    show_default=True,
)
@click.option("--tags", "-t", help="List of tags using comma as separator: tag1,tag2,tag3", type=str, required=False, callback=parse_tags)
def collect_tech_support(ctx: click.Context, tags: Optional[List[str]], output: Path, latest: Optional[int], configure: bool) -> None:
    """Collect scheduled tech-support from EOS devices"""
    asyncio.run(collect_scheduled_show_tech(ctx.obj["inventory"], output, configure, tags=tags, latest=latest))



"""Manage Exception in Inventory module."""


class InventoryRootKeyError(Exception):
    """Error raised when inventory root key is not found."""


class InventoryIncorrectSchema(Exception):
    """Error when user data does not follow ANTA schema."""



#!/usr/bin/env python
# coding: utf-8 -*-

"""
Utils functions to use with anta.cli.get.commands module.
"""

import json
import logging
from pathlib import Path
from typing import Any, Dict, List, Union

import requests
import urllib3
import yaml

from ...inventory import AntaInventory

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

logger = logging.getLogger(__name__)


def get_cv_token(cvp_ip: str, cvp_username: str, cvp_password: str) -> str:
    """Generate AUTH token from CVP using password"""

    # use CVP REST API to generate a token
    URL = f"https://{cvp_ip}/cvpservice/login/authenticate.do"
    payload = json.dumps({"userId": cvp_username, "password": cvp_password})
    headers = {"Content-Type": "application/json", "Accept": "application/json"}

    response = requests.request("POST", URL, headers=headers, data=payload, verify=False, timeout=10)
    return response.json()["sessionId"]


def create_inventory_from_cvp(inv: List[Dict[str, Any]], directory: str, container: str) -> None:
    """
    create an inventory file from Arista CloudVision
    """
    i: Dict[str, Dict[str, Any]] = {AntaInventory.INVENTORY_ROOT_KEY: {"hosts": []}}
    logger.debug(f"Received {len(inv)} device(s) from CVP")
    for dev in inv:
        logger.info(f'   * adding entry for {dev["hostname"]}')
        i[AntaInventory.INVENTORY_ROOT_KEY]["hosts"].append({"host": dev["ipAddress"], "name": dev["hostname"], "tags": [dev["containerName"].lower()]})
    # write the devices IP address in a file
    inv_file = "inventory" if container is None else f"inventory-{container}"
    out_file = f"{directory}/{inv_file}.yml"
    with open(out_file, "w", encoding="UTF-8") as out_fd:
        out_fd.write(yaml.dump(i))
    logger.info(f"Inventory file has been created in {out_file}")


def create_inventory_from_ansible(inventory: Path, output_file: Path, ansible_root: Union[str, None] = None) -> None:
    """
    Create an ANTA inventory from an Ansible inventory YAML file

    Args:
        inventory (str): Ansible Inventory file to read
        output_file (str, optional): ANTA inventory file to generate.
        ansible_root (Union[str, None], optional): Ansible group from where to extract data. Defaults to None.
    """

    def deep_yaml_parsing(data: Dict[str, Any], hosts: Union[None, List[Dict[str, str]]] = None) -> Union[None, List[Dict[str, str]]]:
        """Deep parsing of YAML file to extract hosts and associated IPs"""
        if hosts is None:
            hosts = []
        for key, value in data.items():
            if isinstance(value, dict) and "ansible_host" in value.keys():
                hosts.append({"name": key, "host": value["ansible_host"]})
            elif isinstance(value, dict):
                deep_yaml_parsing(value, hosts)
            else:
                return hosts
        return hosts

    i: Dict[str, Dict[str, Any]] = {AntaInventory.INVENTORY_ROOT_KEY: {"hosts": []}}
    with open(inventory, encoding="utf-8") as inv:
        ansible_inventory = yaml.safe_load(inv)
    if ansible_root not in ansible_inventory.keys():
        logger.error(f"Group {ansible_root} not in ansible inventory {inventory}")
        raise ValueError(f"Group {ansible_root} not in ansible inventory {inventory}")
    if ansible_root is None:
        ansible_hosts = deep_yaml_parsing(ansible_inventory, hosts=[])
    else:
        ansible_hosts = deep_yaml_parsing(ansible_inventory[ansible_root], hosts=[])
    if ansible_hosts is None:
        ansible_hosts = []
    for dev in ansible_hosts:
        logger.info(f'   * adding entry for {dev["name"]}')
        i[AntaInventory.INVENTORY_ROOT_KEY]["hosts"].append({"host": dev["host"], "name": dev["name"]})
    with open(output_file, "w", encoding="UTF-8") as out_fd:
        out_fd.write(yaml.dump(i))
    logger.info(f"Inventory file has been created in {output_file}")



"""
Generic routing test functions
"""

from typing import Optional

from anta.models import AntaCommand, AntaTest


class VerifyRoutingProtocolModel(AntaTest):
    """
    Verifies the configured routing protocol model is the one we expect.
    And if there is no mismatch between the configured and operating routing protocol model.

        model(str): Expected routing protocol model (multi-agent or ribd). Default is multi-agent
    """

    name = "VerifyRoutingProtocolModel"
    description = (
        "Verifies the configured routing protocol model is the expected one and if there is no mismatch between the configured and operating routing protocol model."
    )
    categories = ["routing", "generic"]
    # "revision": 3
    commands = [AntaCommand(command="show ip route summary")]

    @AntaTest.anta_test
    def test(self, model: Optional[str] = "multi-agent") -> None:
        """Run VerifyRoutingProtocolModel validation"""

        if not model:
            self.result.is_skipped("VerifyRoutingProtocolModel was not run as no model was given")
            return
        command_output = self.instance_commands[0].json_output

        configured_model = command_output["protoModelStatus"]["configuredProtoModel"]
        operating_model = command_output["protoModelStatus"]["operatingProtoModel"]
        if configured_model == operating_model == model:
            self.result.is_success()
        else:
            self.result.is_failure(f"routing model is misconfigured: configured: {configured_model} - operating: {operating_model} - expected: {model}")


class VerifyRoutingTableSize(AntaTest):
    """
    Verifies the size of the IP routing table (default VRF).
    Should be between the two provided thresholds.

    Args:
        minimum(int): Expected minimum routing table (default VRF) size.
        maximum(int): Expected maximum routing table (default VRF) size.
    """

    name = "VerifyRoutingTableSize"
    description = "Verifies the size of the IP routing table (default VRF). Should be between the two provided thresholds."
    categories = ["routing", "generic"]
    # "revision": 3
    commands = [AntaCommand(command="show ip route summary")]

    @AntaTest.anta_test
    def test(self, minimum: Optional[int] = None, maximum: Optional[int] = None) -> None:
        """Run VerifyRoutingTableSize validation"""

        if not minimum or not maximum:
            self.result.is_skipped(f"VerifyRoutingTableSize was not run as either minimum {minimum} or maximum {maximum} was not provided")
            return
        if not isinstance(minimum, int) or not isinstance(maximum, int):
            self.result.is_error(f"VerifyRoutingTableSize was not run as either minimum {minimum} or maximum {maximum} is not a valid value (integer)")
            return
        if maximum < minimum:
            self.result.is_error(f"VerifyRoutingTableSize was not run as minimum {minimum} is greate than maximum {maximum}.")
            return

        command_output = self.instance_commands[0].json_output
        total_routes = int(command_output["vrfs"]["default"]["totalRoutes"])
        if minimum <= total_routes <= maximum:
            self.result.is_success()
        else:
            self.result.is_failure(f"routing-table has {total_routes} routes and not between min ({minimum}) and maximum ({maximum})")


class VerifyBFD(AntaTest):
    """
    Verifies there is no BFD peer in down state (all VRF, IPv4 neighbors).
    """

    name = "VerifyBFD"
    description = "Verifies there is no BFD peer in down state (all VRF, IPv4 neighbors)."
    categories = ["routing", "generic"]
    # revision 1 as later revision introduce additional nesting for type
    commands = [AntaCommand(command="show bfd peers", revision=1)]

    @AntaTest.anta_test
    def test(self) -> None:
        """Run VerifyBFD validation"""

        command_output = self.instance_commands[0].json_output

        self.result.is_success()

        for _, vrf_data in command_output["vrfs"].items():
            for _, neighbor_data in vrf_data["ipv4Neighbors"].items():
                for peer, peer_data in neighbor_data["peerStats"].items():
                    if (peer_status := peer_data["status"]) != "up":
                        failure_message = f"bfd state for peer '{peer}' is {peer_status} (expected up)."
                        if (peer_l3intf := peer_data.get("l3intf")) is not None and peer_l3intf != "":
                            failure_message += f" Interface: {peer_l3intf}."
                        self.result.is_failure(failure_message)



"""
Test functions related to the EOS various logging settings

NOTE: 'show logging' does not support json output yet
"""
from __future__ import annotations

import logging
import re
from typing import List, Optional

from anta.models import AntaCommand, AntaTest


def _get_logging_states(logger: logging.Logger, command_output: str) -> str:
    """
    Parse "show logging" output and gets operational logging states used
    in the tests in this module.

    Args:
        command_output: The 'show logging' output
    """
    log_states = command_output.partition("\n\nExternal configuration:")[0]
    logger.debug(f"Device logging states:\n{log_states}")
    return log_states


class VerifyLoggingPersistent(AntaTest):
    """
    Verifies if logging persistent is enabled and logs are saved in flash.

    Expected Results:
        * success: The test will pass if logging persistent is enabled and logs are in flash.
        * failure: The test will fail if logging persistent is disabled or no logs are saved in flash.
    """

    name = "VerifyLoggingPersistent"
    description = "Verifies if logging persistent is enabled and logs are saved in flash."
    categories = ["logging"]
    commands = [
        AntaCommand(command="show logging", ofmt="text"),
        AntaCommand(command="dir flash:/persist/messages", ofmt="text"),
    ]

    @AntaTest.anta_test
    def test(self) -> None:
        """
        Run VerifyLoggingPersistent validation.
        """
        self.result.is_success()

        log_output = self.instance_commands[0].text_output
        dir_flash_output = self.instance_commands[1].text_output

        if "Persistent logging: disabled" in _get_logging_states(self.logger, log_output):
            self.result.is_failure("Persistent logging is disabled")
            return

        pattern = r"-rw-\s+(\d+)"
        persist_logs = re.search(pattern, dir_flash_output)

        if not persist_logs or int(persist_logs.group(1)) == 0:
            self.result.is_failure("No persistent logs are saved in flash")


class VerifyLoggingSourceIntf(AntaTest):
    """
    Verifies logging source-interface for a specified VRF.

    Expected Results:
        * success: The test will pass if the provided logging source-interface is configured in the specified VRF.
        * failure: The test will fail if the provided logging source-interface is NOT configured in the specified VRF.
        * skipped: The test will be skipped if source-interface or VRF is not provided.
    """

    name = "VerifyLoggingSourceInt"
    description = "Verifies logging source-interface for a specified VRF."
    categories = ["logging"]
    commands = [AntaCommand(command="show logging", ofmt="text")]

    @AntaTest.anta_test
    def test(self, intf: Optional[str] = None, vrf: str = "default") -> None:
        """
        Run VerifyLoggingSrcDst validation.

        Args:
            intf: Source-interface to use as source IP of log messages.
            vrf: The name of the VRF to transport log messages. Defaults to 'default'.
        """
        if not intf or not vrf:
            self.result.is_skipped(f"{self.__class__.name} did not run because intf or vrf was not supplied")
            return

        output = self.instance_commands[0].text_output

        pattern = rf"Logging source-interface '{intf}'.*VRF {vrf}"

        if re.search(pattern, _get_logging_states(self.logger, output)):
            self.result.is_success()
        else:
            self.result.is_failure(f"Source-interface '{intf}' is not configured in VRF {vrf}")


class VerifyLoggingHosts(AntaTest):
    """
    Verifies logging hosts (syslog servers) for a specified VRF.

    Expected Results:
        * success: The test will pass if the provided syslog servers are configured in the specified VRF.
        * failure: The test will fail if the provided syslog servers are NOT configured in the specified VRF.
        * skipped: The test will be skipped if syslog servers or VRF are not provided.
    """

    name = "VerifyLoggingHosts"
    description = "Verifies logging hosts (syslog servers) for a specified VRF."
    categories = ["logging"]
    commands = [AntaCommand(command="show logging", ofmt="text")]

    @AntaTest.anta_test
    def test(self, hosts: Optional[List[str]] = None, vrf: str = "default") -> None:
        """
        Run VerifyLoggingHosts validation.

        Args:
            hosts: List of hosts (syslog servers) IP addresses.
            vrf: The name of the VRF to transport log messages. Defaults to 'default'.
        """
        if not hosts or not vrf:
            self.result.is_skipped(f"{self.__class__.name} did not run because hosts or vrf were not supplied")
            return

        output = self.instance_commands[0].text_output

        not_configured = []

        for host in hosts:
            pattern = rf"Logging to '{host}'.*VRF {vrf}"
            if not re.search(pattern, _get_logging_states(self.logger, output)):
                not_configured.append(host)

        if not not_configured:
            self.result.is_success()
        else:
            self.result.is_failure(f"Syslog servers {not_configured} are not configured in VRF {vrf}")


class VerifyLoggingLogsGeneration(AntaTest):
    """
    Verifies if logs are generated.

    Expected Results:
        * success: The test will pass if logs are generated.
        * failure: The test will fail if logs are NOT generated.
    """

    name = "VerifyLoggingLogsGeneration"
    description = "Verifies if logs are generated."
    categories = ["logging"]
    commands = [
        AntaCommand(command="send log level informational message ANTA VerifyLoggingLogsGeneration validation"),
        AntaCommand(command="show logging informational last 30 seconds | grep ANTA", ofmt="text"),
    ]

    @AntaTest.anta_test
    def test(self) -> None:
        """
        Run VerifyLoggingLogs validation.
        """
        log_pattern = r"ANTA VerifyLoggingLogsGeneration validation"

        output = self.instance_commands[1].text_output
        lines = output.strip().split("\n")[::-1]

        for line in lines:
            if re.search(log_pattern, line):
                self.result.is_success()
                return

        self.result.is_failure("Logs are not generated")


class VerifyLoggingHostname(AntaTest):
    """
    Verifies if logs are generated with the device FQDN.

    Expected Results:
        * success: The test will pass if logs are generated with the device FQDN.
        * failure: The test will fail if logs are NOT generated with the device FQDN.
    """

    name = "VerifyLoggingHostname"
    description = "Verifies if logs are generated with the device FQDN."
    categories = ["logging"]
    commands = [
        AntaCommand(command="show hostname"),
        AntaCommand(command="send log level informational message ANTA VerifyLoggingHostname validation"),
        AntaCommand(command="show logging informational last 30 seconds | grep ANTA", ofmt="text"),
    ]

    @AntaTest.anta_test
    def test(self) -> None:
        """
        Run VerifyLoggingHostname validation.
        """
        output_hostname = self.instance_commands[0].json_output
        output_logging = self.instance_commands[2].text_output
        fqdn = output_hostname["fqdn"]
        lines = output_logging.strip().split("\n")[::-1]

        log_pattern = r"ANTA VerifyLoggingHostname validation"

        last_line_with_pattern = ""
        for line in lines:
            if re.search(log_pattern, line):
                last_line_with_pattern = line
                break

        if fqdn in last_line_with_pattern:
            self.result.is_success()
        else:
            self.result.is_failure("Logs are not generated with the device FQDN")


class VerifyLoggingTimestamp(AntaTest):
    """
    Verifies if logs are generated with the approprate timestamp.

    Expected Results:
        * success: The test will pass if logs are generated with the appropriated timestamp.
        * failure: The test will fail if logs are NOT generated with the appropriated timestamp.
    """

    name = "VerifyLoggingTimestamp"
    description = "Verifies if logs are generated with the appropriate timestamp."
    categories = ["logging"]
    commands = [
        AntaCommand(command="send log level informational message ANTA VerifyLoggingTimestamp validation"),
        AntaCommand(command="show logging informational last 30 seconds | grep ANTA", ofmt="text"),
    ]

    @AntaTest.anta_test
    def test(self) -> None:
        """
        Run VerifyLoggingTimestamp validation.
        """
        log_pattern = r"ANTA VerifyLoggingTimestamp validation"
        timestamp_pattern = r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{6}-\d{2}:\d{2}"

        output = self.instance_commands[1].text_output

        lines = output.strip().split("\n")[::-1]

        last_line_with_pattern = ""
        for line in lines:
            if re.search(log_pattern, line):
                last_line_with_pattern = line
                break

        if re.search(timestamp_pattern, last_line_with_pattern):
            self.result.is_success()
        else:
            self.result.is_failure("Logs are not generated with the appropriate timestamp format")


class VerifyLoggingAccounting(AntaTest):
    """
    Verifies if AAA accounting logs are generated.

    Expected Results:
        * success: The test will pass if AAA accounting logs are generated.
        * failure: The test will fail if AAA accounting logs are NOT generated.
    """

    name = "VerifyLoggingAccounting"
    description = "Verifies if AAA accounting logs are generated."
    categories = ["logging"]
    commands = [AntaCommand(command="show aaa accounting logs | tail", ofmt="text")]

    @AntaTest.anta_test
    def test(self) -> None:
        """
        Run VerifyLoggingAccountingvalidation.
        """
        pattern = r"cmd=show aaa accounting logs"
        output = self.instance_commands[0].text_output

        if re.search(pattern, output):
            self.result.is_success()
        else:
            self.result.is_failure("AAA accounting logs are not generated")
